-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- Abschnitt 2.1.4 Latches
--
-- Beschreibung des Latch-Blockspeichers (Abb. 2.23)
-- in zwei Versionen
-- * "Synth": Synthesebeschreibung aus dem Buch
-- * "Sim": Simulationsmodell mit Berücksichtigung des
--   Zeitverhaltens und mit Berechnung der Gültigkeit
--------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;

entity LatchBS is
 port(E: in std_logic;
      a: in std_logic_vector(1 downto 0);
      x: in std_logic_vector(3 downto 0);
      y0, y1, y2, y3: buffer std_logic_vector(3 downto 0));
end entity;

architecture Synth of LatchBS is
 signal s, q: std_logic_vector(3 downto 0);
begin

Dec:process(a)
 begin
  case a is
   when "00"   => s <= "0001";
   when "01"   => s <= "0010";
   when "10"   => s <= "0100";
   when others => s <= "1000";
  end case;
 end process;

Und:process(s, E)
 begin
  if E='1' then q <= s;
  else q <= "0000";
  end if;
 end process;

Latch0:process(x, q(0))
 begin
  if q(0)='1' then y0<=x; end if;
 end process;

Latch1:process(x, q(1))
 begin
  if q(1)='1' then y1<=x; end if;
 end process;

Latch2:process(x, q(2))
 begin
  if q(2)='1' then y2<=x; end if;
 end process;

Latch3:process(x, q(3))
 begin
  if q(3)='1' then y3<=x; end if;
 end process;
end architecture;

--===============================================
-- Simulationsmodell mit Verzögerungszeiten und 
-- Berechnung der Signalgültigkeit
architecture Sim of LatchBS is
 signal s, q: std_logic_vector(3 downto 0);
 constant th1: delay_length:=  50 ps;
 constant td1: delay_length:= 150 ps;
 constant th2: delay_length:=  40 ps;
 constant td2: delay_length:= 100 ps;
begin

-- Der Decoder wird ganz normal mit pessimistischen Halte-
-- und Verzögerungszeiten simuliert
Dec:process(a)
 begin
  case a is
   when "00" => s <= "XXXX" after th1, "0001" after td1;
   when "01" => s <= "XXXX" after th1, "0010" after td1;
   when "10" => s <= "XXXX" after th1, "0100" after td1;
   when "11" => s <= "XXXX" after th1, "1000" after td1;
   when others => s <= "XXXX" after th1;
  end case;
 end process;

-- Die UND-Gatter müssen so modelliert werden, dass, 
-- wenn s(i) oder E stabil null ist, der Wert der 
-- UND-Verknüpfungen nicht nach einer Aenderung
-- von E oder s auf ungültig gesetzt wird
-- 
 process (E, s)
 begin
  if q(0)/=(E and s(0)) then q(0) <='X' after th2, E and s(0) after td2; end if;
  if q(1)/=(E and s(1)) then q(1) <='X' after th2, E and s(1) after td2; end if;
  if q(2)/=(E and s(2)) then q(2) <='X' after th2, E and s(2) after td2; end if;
  if q(3)/=(E and s(3)) then q(3) <='X' after th2, E and s(3) after td2; end if;
 end process;

-- Das Latchmodell ist aus den vorherigen Beispielen übernommen
Latch0:entity work.Latch port map (E=>q(0), x=>x, y=>y0);
Latch1:entity work.Latch port map (E=>q(1), x=>x, y=>y1);
Latch2:entity work.Latch port map (E=>q(2), x=>x, y=>y2);
Latch3:entity work.Latch port map (E=>q(3), x=>x, y=>y3);

end architecture;
