-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- 3.4.6 Schrittweise Modellentwicklung -- Beispiel UART
--
-- Verbessertes UART-Funktionsmodell 
-- a) taktsynchron und asynchroner Initialisierung
-- b) Beschreibung als Automatenbeschreibung
---------------------------------------------------------------------


-- Vorspann
library ieee;
use ieee.std_logic_1164.all;

use std.textio.all;

library Tuc;
use Tuc.Eingabe.all;
use Tuc.Ausgabe.all; 
use Tuc.Numeric_Sim.all;


package UART2_pack is

 subtype tByte is std_logic_vector(7 downto 0);
 constant XX: tByte:=(others=>'X');
-- ===================================================================
-- taktsynchron und asynchroner Initialisierung
-- Prozedur zur Beschreibung des Senders
 procedure Sender(signal T, I, Start: std_logic; signal x: tByte;
   signal TxD, busy: out std_logic);

-- Prozedur zur Beschreibung des Empfängers
 procedure Empfaenger(signal T, I, RxD: std_logic; signal y:
     out tByte; signal Ready, Err: out std_logic);

-- ===================================================================
-- Datentypen und Funktionen für die synthesefähige Beschreibung

-- Datentypen für den Zustand des Senders
 type tZTyp is (zStopp, zStart, zDaten, zParitaet);
 type tZustand is record 
  z:   tZTyp;
  idx: tUnsigned(2 downto 0);
  P:   std_logic;
 end record;

-- Str-Funktionen für den Zustand
 function str(z: tZTyp) return string;
 function str(z: tZustand) return string;

-- Uebergangsfunktion
 function fz_Sender(Start:std_logic; x: tByte; z_in: tZustand) return tZustand;

-- Ausgabefunktionen
 function f_TxD(x: tByte; z: tZustand) return std_logic;
 function f_Busy(z: tZustand) return std_logic;
end package;


package body UART2_pack is
--===================================================================
--Prozedur zur Beschreibung des Senders 
 procedure Sender(signal T, I, Start: std_logic; signal x: tByte;
   signal TxD, busy: out std_logic) is
  variable P: std_logic;
 begin
  S0: loop
   wait until falling_edge(I);
   S1: loop
    TxD<='1'; Busy<='0';
    wait until (rising_edge(T) and Start='1') or I='1'; 
    exit S1 when I='1';
    TxD<='0'; Busy<='1'; P:='0';
    wait until rising_edge(T)  or I='1'; exit S1 when I='1';
    S2: for idx in x'reverse_range loop
     TxD<=x(idx); P:=P xor x(idx);
     wait until rising_edge(T)  or I='1'; exit S1 when I='1';
    end loop;
    TxD<=P; 
    wait until rising_edge(T)  or I='1'; exit S1 when I='1';
   end loop;
  end loop;
 end procedure;

--===================================================================
 procedure Warte(signal T, I: std_logic; ct: inout tUnsigned(3 downto 0)) is
 begin
  loop
   ct:=ct+"1"; 
   wait until rising_edge(T) or I='1';
   exit when I='1' or ct="0000";
  end loop; 
 end procedure;
---------------------------------------------------------------------
-- Prozedur zur Beschreibung des Empfängers
 procedure Empfaenger(signal T, I, RxD: std_logic; signal y:
     out tByte; signal Ready, Err: out std_logic) is
  variable P: std_logic;
  variable ct: tUnsigned(3 downto 0);
 begin
  S0: loop
   wait until rising_edge(I);
   Ready<='0'; Err<='0'; 
   S1: loop
    wait until (rising_edge(T) and RxD='0') or (I='1');
    exit S1 when I='1';
    Ready<='0'; Err<='0'; P:='0';
    ct:="1000";
    Warte(T, I, ct); exit S1 when I='1';
    Warte(T, I, ct); exit S1 when I='1';
    S2: for idx in y'reverse_range loop 
     y(idx)<=RxD;
     P:=P xor RxD;
     Warte(T, I, ct); exit S1 when I='1';
    end loop;
    Ready<='1';
    if P/=RxD then Err<='1'; end if;
    Warte(T, I, ct); exit S1 when I='1';
    if RxD/='1' then Err<='1'; end if;
    ct:="1000"; Warte(T, I, ct); 
   end loop;
  end loop;
 end procedure;

--===================================================================
-- Str-Funktion für den Zustand
 function str(z: tZTyp) return string is
 begin
  case z is
   when zStopp    => return "Sto";
   when zStart    => return "Sta";
   when zDaten    => return "Dat";
   when zParitaet => return "Par";
   when others    => return "XXX";
  end case;
 end function;

 function str(z: tZustand) return string is
 begin
  return str(z.z) & ':' & str(z.idx) & ':' & str(z.P);
 end function;

-- Uebergangsfunktion
 function fz_Sender(Start:std_logic; x: tByte; z_in: tZustand) return tZustand is
  variable z: tZustand:=z_in;
 begin
  case z.z is
   when zStopp => z.idx:="000"; z.P := '0';
    if Start='1' then z.z:=zStart; end if;
   when zStart =>  z.idx:="000"; z.P:='0'; z.z:=zDaten;
   when zDaten => z.idx:=z.idx+"1"; z.P:=z.P xor x(int(z.idx)); 
    if z.idx="000" then z.z:=zParitaet; end if;
   when zParitaet => z.idx:="000"; z.P := '0'; z.z:=zStopp;
  end case;
  return z;
 end function;

-- Ausgabefunktion
 function f_TxD(x: tByte; z: tZustand) return std_logic is
 begin
  case z.z is
   when zStart  => return '0';
   when zDaten =>  return x(int(z.idx)); 
   when zParitaet => return z.P;
   when others => return '1';
  end case;
 end function;

 function f_Busy(z: tZustand) return std_logic is
 begin
  if z.z=zStopp then return '1'; else return '0'; end if;
 end function;


end package body;
