-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- 3.4.7 Entwicklung eines CORDIC-Rechenwerks
--
-- Package für ein CORDIC-Rechenwerk mit Festkommazahlen
---------------------------------------------------------------------


library ieee;
use ieee.math_real.cos;
use ieee.math_real.atan;

library Tuc;
use Tuc.Numeric_Sim.all;
use Tuc.Ausgabe.all;

package cordic2_pack is

-- Datendarstellung durch ein Festkommaformat
-- die Kommaposition ist hinter der zweiten Bitstelle,
-- erkennbar an der Multiplikation von 2^14 bei der
-- Konvertierung in eine Gleitkommazahl

 subtype tDaten is tSigned(15 downto 0); 
 type tDatenArray is array (natural range <>) of tDaten;

 function str_real(x: tDaten) return string;
 function to_tDat(w: real) return tDaten;

--========================================================
-- Funktionen zur Berechnung der Konstantentabelle und der 
-- Konstanten SCS
 function fAtanTab(N: positive) return tDatenArray;
 function fSCS(Tab: tDatenArray) return tDaten;
--========================================================
-- Umdefinition der Multiplikation mit Rückgabe des Produktes
-- mit dem Typ tDaten und der gedachten Kommaposition nach dem
-- zweiten Bit
 function mult(a, b: tDaten) return tDaten;

end package;

package body cordic2_pack is
--========================================================
-- Konvertierfunktionen
 function str_real(x: tDaten) return string is
 begin
  return str(real(int(x))/(2.0**14), 4);
 end function;

 function to_tDat(w: real) return tDaten is
 begin
  assert w>-2.0 and w<2.0 report "Wertebereichsverletzung" 
   severity failure;
  return to_tSigned(integer(w*2.0**14), 16);
 end function;

 function fAtanTab(N: positive) return tDatenArray is
 variable Tab: tDatenArray(0 to N-1);
 begin
  for idx in Tab'range loop
   Tab(idx):=to_tDat(atan(2.0 ** (-idx)));
  end loop;
  return Tab;
 end function;
--========================================================
-- Funktionen zur Berechnung der Konstantentabelle und der 
-- Konstanten SCS

 function fSCS(Tab: tDatenArray) return tDaten is
  variable w: real:=1.0;
 begin
  for idx in Tab'range loop
   w:=w*cos(atan(2.0 ** (-idx)));
  end loop;
  return to_tDat(w);
 end function;
--========================================================
-- Umdefinition der Multiplikation mit Rückgabe des Produktes
-- mit dem Typ tDaten und der gedachten Kommaposition nach dem
-- zweiten Bit

 function mult(a, b: tDaten) return tDaten is
  constant y: tSigned(2*tDaten'length-1 downto 0):=a*b;
 begin
  return y(y'high-2 downto y'high-tDaten'length-1);
 end function;

end package body;
