-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- 5.2 FIR-Filter mit Blockspeichern, Untersabschn. 5.2.3 und 5.2.4
-- 
--
-- Package für das erste Simulationsmodell des FIR-Filters
--
-- Zahlentypen für Adressen und Daten: tAdr, tDat
-- Datentyp für die Blockspeicher:     tMem
-- Verbund für alle Pipeline-Register: tPipeline
-- 
-- Initialisierungs- und Verarbeitungsschritt der Pipline:
--  procedure InitStep(Eingabewert, xRAM, Pipeline);
--  procedure NormalStep(Eingabe, cROM, xRAM, Pipeline);
--
-- Str-Funktion und Einleseprozedur aus einer Datei für tDat 
--  function str(Datenwert) => formatierte_Textdarstellung
--  procedure read(Eingabedatei, Eingabevariable_vom_Typ_tDat);
--
-- Erzeugung einer Textzeile des Schaltungszustands
--  constant StatusKopfText
--  function str(Pipeline-Zustand, Datenspeicher) => Textzeile
---------------------------------------------------------------------

use STD.TEXTIO.TEXT;

library Tuc;
use Tuc.Ausgabe.all;
use Tuc.Eingabe.all;

package FIR1_pack is

 constant M: positive:=4;                -- Filtergrad
 constant max_abs: real:=1000.0;	 -- Wertebereichsgrenzen
 subtype tAdr is natural range 0 to M-1;
 type tDat is range -max_abs to max_abs;
 type tMem is array (natural range <>) of tDat;

 type tPipeline is record
  cAdr, xAdr: tAdr;
  cDat, xDat, Prod, Akku: tDat;
 end record;

 procedure InitStep(x: tDat; signal xRAM: inout tMem; signal pp: inout tPipeline);
 procedure NormalStep(x: tDat; cROM: tMem; signal xRAM: inout tMem; signal pp: inout tPipeline);


 -- Str-Funktion und Einleseprozedur aus einer Datei für tDat 
 function str(x: tDat; n: natural:=1) return string;
 procedure read(file f: TEXT; w: out tDat);

 -- Funktion zur Erzeugung einer Textzeile mit dem Pipeline-Zustand
 -- und eine Zeichenkette als Tabellenkopf 
 constant StatusKopfText: string :=
  "|cAdr|xAdr||   cDat |   xDat ||  Prod  |  Akku  ||   r(0)    r(1)    r(2)    r(3) |";
 function str(pp: tPipeline; xRAM: tMem) return string;

end package;



package body FIR1_pack is
------------------------------------------------------------------------
-- Ueberladen der Addition und und der Multiplikation für den Typ tDat
 function "+"(a, b: tDat) return tDat is
  constant sum: real:= real(a)+real(b);
 begin
  if a=tDat'LOW or b=tDat'LOW or (abs sum)>max_abs then return tDat'LOW;
  else return tDat(sum);
  end if;
 end function; 

 function "*"(a, b: tDat) return tDat is
  constant prod: real:= real(a)*real(b);
 begin
  if a=tDat'LOW or b=tDat'LOW or (abs prod)>max_abs then return tDat'LOW;
  else return tDat(prod);
  end if;
 end function; 

 -------------------------------------------------------------------------
 -- Initialisierungsschritt
 procedure InitStep(x: tDat; signal xRAM: inout tMem; signal pp: inout tPipeline) is
 begin
  pp.cAdr <= (pp.cAdr + 1) mod M;
  pp.xAdr <= pp.xAdr - 1;
  xRAM(pp.xAdr) <= x;
 end procedure;

 -- Verarbeitungsschritt
 procedure NormalStep(x: tDat; cROM: tMem; signal xRAM: inout tMem; signal pp: inout tPipeline) is
 begin
  pp.cAdr<=(pp.cAdr+1) mod M;
  if pp.cAdr=0 then 
   xRAM(pp.xAdr) <= x;
   pp.xDat<=x;
  else 
   pp.xDat <= xRAM(pp.xAdr);   
  end if;
  pp.cDat <= cROM(pp.cAdr);
  pp.Prod <= pp.cDat * pp.xDat;
  if pp.cAdr=2 then pp.Akku<=pp.Prod; 
  else pp.Akku<=pp.Akku+pp.Prod; 
  end if;
  if pp.cAdr/=M-1 then pp.xAdr<=(pp.xAdr+1) mod M; end if;
 end procedure;

 -------------------------------------------------------------------------
 -- Erzeugung einer kompakten, formatierten Textdarstellung für die Daten
 function str(x: tDat; n: natural:=1) return string is
 begin
  if x=tDat'LOW then return rechts("XX  ", 8);
  else return rechts(str(real(x),n),8);
  end if;
 end function;

-- Prozedur zum Einlesen von Daten aus einer Datei
 procedure read(file f: TEXT; w: out tDat) is
  variable pstr: tPString;
  variable y: REAL;
 begin
  read(f, pstr);
   --write(str(pstr));
  get(pstr, y, -max_abs, max_abs);
  assert pstr.Status=ok report str(pstr.err_msg) severity failure;
  w := tDat(y);
 end procedure; 

 -------------------------------------------------------------------------
 -- Erzeugung einer Textzeile des Schaltungszustands der Form:
 -- |cAdr|xAdr|| cDat| xDat|| Prod| Akku|| r(0) r(1) r(2) r(3)|
 function str(pp: tPipeline; xRAM: tMem) return string is
  variable s: tString;
 begin
  assign(s, "|" & rechts(str(pp.cAdr),4) & "|" & rechts(str(pp.xAdr),4) &
  "||" & str(pp.cDat, 4) & "|" & str(pp.xDat) &  "||" & str(pp.Prod) 
  & "|" & str(pp.Akku) & "||");
  for idx in 0 to M-2 loop
   append(s, str(xRAM(idx)));
  end loop; 
  append(s, str(xRAM(M-1))&'|');
  return str(s);
 end function;
 -------------------------------------------------------------------------
end package body;
