-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- 5.2 FIR-Filter mit Blockspeichern, ab Untersabschn. 5.2.5
-- 
--
-- Package für FIR-Filter mit Wertedarstellung durch Bitvektoren 
-- und Ein- und Ausgabe über Signale
--
-- Neudefinition von  tAdr, tDat als Bitvektortypen zur Darstellung vor-
-- zeichenbehafteter Zahlen; grösserte Akku-Bitbreite und Umdefinition der 
-- Multiplikation, um Zahlenbereichüberläufe auszuschliessen
--
-- Beschreibungsschablone Steuerautomat; Erweiterung von tPipeline um ein 
-- Zustandsregister
-- 
-- Einzelbit-Signale:  T (Takt), I (Init), RD (Lesenbestätigung), WR (Schreibbestätigung)
-- Datensignale:       x (Eingabe), y (Ausgabe, identisch mit pp.Akku)
-- Blockspeicher:      cROM (Koeffizientenspeicher/Konstante), xRAM (Datenspeicher/Signal)
-- Signalverbund:      pp
--
-- * Uebergangsfunktion 
--   procedure PipeStep(x, cROM, xRAM, pp);
--
-- * Ausgabefunktionen zur Erzeugung der Bestätigungssignale für die 
--   Ein- und Ausgabe
--   function RdGnt(pp) => RD
--   function WrGnt(pp) => WR
-- 
-- Funktionsbeschreibung der Testdatenquelle
--  procedure Testdatenquelle(Eingabedateiname, RD, T, I, x, tp: DELAY_LENGTH:=10 ns);
--
-- Prozedur zur Protokollierung der Testausgaben
--  procedure Testausgabe(T, RD, WR, pp, x, xRAM)
--
-- Konvertierfunktion function to_tKoeff(w) => Filterkoeffizient
-- zur Umwandlung eines Gleitkommawertes abs(w)<1.0 in einen Filterkoeffizienten vom 
-- Typ tSigned mit dem Wert to_tSigned(integer(w * 2**tDat'LENGTH))
---------------------------------------------------------------------

use STD.TEXTIO;
library IEEE;
use IEEE.STD_LOGIC_1164.all;

library Tuc;
use Tuc.Ausgabe.all;
use Tuc.Eingabe.all;
use Tuc.Numeric_Sim.all;

package FIR2_pack is


 constant M: positive:=4;                -- Filtergrad
 subtype tAdr is tUnsigned(1 downto 0); 
 subtype tDat is tSigned(15 downto 0);
 subtype tAkku is tSigned(17 downto 0);
 type tMem is array (natural range <>) of tDat;
 constant xxx: tDat:=x"4456";
 type tZustand is (XX, Init, Startzyklus, Normalzyklus);

 type tPipeline is record
  cAdr, xAdr: tAdr;
  cDat, xDat: tDat;
  Akku, Prod: tAkku;
  Zustand: tZustand;
 end record;

 -- Uebergangsfunktion
 procedure PipeStep(x: tDat; cROM: tMem; signal xRAM: inout tMem; signal pp: inout tPipeline);

 -- Ausgabefunktionen
 function RdGnt(pp: tPipeline) return std_logic;
 function WrGnt(pp: tPipeline) return std_logic;

 -- Testdatenquelle
 procedure Testdatenquelle(Dateiname: string; signal RD: std_logic; signal T, I: out std_logic;
  signal x: out tDat; tp: DELAY_LENGTH:=10 ns);

 -- Protokollierung der Testausgaben
 procedure Testausgabe(signal T, RD, WR: std_logic; signal pp: tPipeline; 
   signal x: tDat; signal xRAM: tMem);


 -- Hilfsfunktionen für den Testrahmen
 --  Konvertierung von Gleitkommazahlen in Filterkoeffizienten
 function to_tKoeff(x: REAL) return tDat;


 function mult(a, b: tDat) return tAkku;
 function str_dat(x: tSigned) return string;
 function str_coeff(c: tDat) return string;

end package;



package body FIR2_pack is
--==========================================================================================
-- nicht exportierte Hilfsfunktionen der Schaltungsbeschreibung
--  Lesefunktion für Blockspeicher
 function read(Mem: tMem; adr: tAdr) return tDat is
  constant cXX: tDat:= (others=>'X');
  variable idx: NATURAL;
 begin
  if is_x(adr) then return cXX;
  else
   idx:= int(adr);
   if idx<Mem'low or idx>Mem'high then return cXX;
   else return Mem(idx);
   end if;
  end if;
 end function;
-------------------------------------------------------------------------------------------
-- Schreib-Prozedur für Blockspeicher
 procedure write(signal Mem: inout tMem; adr: tAdr; x: tDat) is
  constant cXX: tDat:= (others=>'X');
  variable idx: NATURAL;
 begin
  if is_x(adr) then 
   for idx in Mem'range loop
    Mem(idx)<=cXX;
   end loop;
  else
   idx:= int(adr);
   if idx>=Mem'low and idx<=Mem'high then 
    Mem(idx) <= x;
   end if;
  end if;
 end procedure;
-------------------------------------------------------------------------------------------
--  Multiplikation mit Rückgabe der höherwertigen Produkthälfte
 function mult(a, b: tDat) return tAkku is
  constant prod: tSigned(2*tDat'LENGTH-1 downto 0):= a*b;
 begin
  return prod(prod'HIGH) & prod(prod'HIGH downto tDat'HIGH);
 end function; 
--==========================================================================================
-- exportierte Unterprogramme zur Schaltungsbeschreibung
-- * Uebergangsfunktion
 procedure PipeStep(x: tDat; cROM: tMem; signal xRAM: inout tMem; signal pp: inout tPipeline) is
 begin
  case pp.Zustand is
   when XX => null;
   when Init =>
    pp.cAdr<=pp.cAdr+"1";
    pp.xAdr <= pp.xAdr - "1";
    --write(xRAM, pp.xAdr, x);
    if pp.cAdr=to_tUnsigned(M-1, tAdr'LENGTH) then pp.Zustand <= Startzyklus; end if;
   when others =>
    pp.cAdr<=pp.cAdr+"1";
    if pp.cAdr=to_tUnsigned(M-1, tAdr'LENGTH) then pp.Zustand<=Normalzyklus; end if;
    if pp.cAdr/=to_tUnsigned(M-1, tAdr'LENGTH) then pp.xAdr <= pp.xAdr + "1"; end if;
    if pp.cAdr="00" then 
     write(xRAM, pp.xAdr, x);
     pp.xDat<=x;
    else 
     pp.xDat <= read(xRAM, pp.xAdr);   
    end if;
    pp.cDat <= read(cROM, pp.cAdr);
    pp.Prod <= mult(pp.cDat, pp.xDat);
    if pp.cAdr=to_tUnsigned(2, tAdr'LENGTH) then pp.Akku<=pp.Prod; 
    else pp.Akku<=pp.Akku + pp.Prod; 
    end if;
  end case;
 end procedure;
 -------------------------------------------------------------------------------------------
 -- * Ausgabefunktion für des Lesebestätigungssignal
 function RdGnt(pp: tPipeline) return std_logic is
 begin
  if pp.Zustand=Init or pp.cAdr=to_tUnsigned(0, tAdr'LENGTH) then
   return '1'; 
  else return '0';
  end if;
 end function;
 -------------------------------------------------------------------------------------------
 -- * Ausgabefunktion fur des Schreibbestätigungssignal
 function WrGnt(pp: tPipeline) return std_logic is
 begin
  if pp.Zustand=Normalzyklus and pp.cAdr=to_tUnsigned(2, tAdr'LENGTH) then
   return '1'; 
  else return '0';
  end if;
 end function;

--==========================================================================================
 -- Simulationsmodell der Datenquelle
 procedure Testdatenquelle(Dateiname: string; signal RD: std_logic; signal T, I: out std_logic;
  signal x: out tDat; tp: DELAY_LENGTH:=10 ns) is
  variable pstr: tPString;
  variable ZNr: natural;
  variable r: real;
  file f: TEXTIO.TEXT open READ_MODE is Dateiname;
 begin
  I <= '1'; T<='0'; wait for tp; I<='0' after tp/10;
  loop
   T <= '1', '0' after tp/2;
   if RD='1' then 
    if TEXTIO.ENDFILE(f) then wait; end if;
    read(f, pstr);
    get(pstr, r, -1000.0, 1000.0);
    if pstr.Status/=ok then
     write("Zeile " & str(znr) & ":" & str(pstr.err_msg));
     wait;
    else x <= to_tSigned(integer(r*(2.0**5)), x'LENGTH) after tp/3; ZNr := ZNr + 1;
    end if;
   end if;
   wait for tp;
  end loop;
 end procedure;

--==========================================================================================
-- Beschreibung der Testausgabe
--  nicht exportierte Hilffunktionen
--   Erzeugung einer formatierten Textdarstellung für Datenwerte
 function str_dat(x: tSigned) return string is
 begin
  if is_x(x) then return "     XX";
  else return rechts(str(real(int(x))/(2.0**5), 1),7);
  end if;
 end function;

 function str_coeff(c: tDat) return string is
 begin
  if is_x(c) then return "     XX";
  else return rechts(str(real(int(c))/(2.0**15), 4),7);
  end if;
 end function;

 -------------------------------------------------------------------------------------------
 --  Darstellung des Zustands durch eine Zeichenkette der Länge eins
 function str(z: tZustand) return STRING is
 begin
  case z is
   when XX          => return "X";
   when Init        => return "I";
   when Startzyklus => return "S";
   when others      => return "N";
  end case;
 end function;
--------------------------------------------------------------------------------------------
--   Prozedur für die Protokollierung der Testausgaben
 procedure Testausgabe(signal T, RD, WR: std_logic; signal pp: tPipeline; 
   signal x: tDat; signal xRAM: tMem) is
  variable s: tString;
 begin
  -- Ausgabe der Kopfzeile
  write("  Zeit  |Eingabe|Z|cAdr|xAdr||  cDat |  xDat ||  Prod |  Akku ||   r(0)   r(1)   r(2)   r(3)|Ausgabe");
  loop
   --warte auf jede steigende Taktflanke
   wait until rising_edge(T);
   -----------------------------------------------------------------------------------------
   -- Erzeuge einen Ausgabestring mit dem Format der Kopfzeile und den aktuellen Signalwerten
   assign(s, rechts(str(NOW, '1'),8));
   if rd='1' then append(s, '|' & str_dat(x)); else append(s, "|       "); end if;
   append(s, "|" &str(pp.Zustand) & '|' & rechts(str(pp.cAdr, 'd'),4) & "|" & rechts(str(pp.xAdr, 'd'),4) &
  "||" & str_coeff(pp.cDat) & "|" & str_dat(pp.xDat) &  "||" & str_dat(pp.Prod) 
  & "|" & str_dat(pp.Akku) & "||");
   for idx in 0 to M-2 loop
    append(s, str_dat(xRAM(idx)));
   end loop; 
   append(s, str_dat(xRAM(M-1))&'|');
   if wr='1' then append(s,str_dat(pp.Akku)); end if;
   ------------ Zeile Schreiben ------------------------------------------------------------
   write(s);
  end loop;
 end procedure;
--==========================================================================================
 -- Hilfsfunktionen für den Testrahmen: Konvertierung von Gleitkommazahlen 
 -- in Filterkoeffizienten
 function to_tKoeff(x: REAL) return tDat is
 begin
  if abs(x) <=1.0-1.0/(2.0**15) then
   return to_tSigned(INTEGER(2.0**(tDat'LENGTH-1) * x), tDat'LENGTH);
  else return (others=>'X');
  end if;
 end function;
--==========================================================================================


end package body;
