-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- 5.4 Entwurf eines RISC-Prozessors
--
-- Package mit den Textverarbeitungsfunktionen des Simulationsmodells
-- eines RISC-Prozessors
--
-- Konvertierfunktionen in Textdarstellungen fuer die Simulation des RISC-Prozessors
--
-- str_RegAdr(tAReg)   => string  -- Registeradressen
-- str_Opcode(tOpcode) => string  -- Operationscode
-- str_Bed(x: tBed)    => string  -- Sprungbedingung
-- str_Instr(tInstr)   => string  -- Befehlswort
-- str(tCtrlTyp)       => string  -- Operationstyp
-- str(tErgebnis)      => string  -- Rechenwerkergebnis (Wert + Flags)
--
-- * Erzeugung einer Kopf und einer Datenzeile fuer den Prozessorstatus
-- * Ausgabe des Befehlsspeicherinhalts in Tabellenform (4 Spalten zu je 20 Zeichen)
-- * Ausgabe des Datenspeicherinhaltes in Tabellenform (8 Spalten zu je 10 Zeichen)
----------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
library Tuc;
use Tuc.Ausgabe.all;
use Tuc.Numeric_Sim.all;
use work.RISC_pack.all;

package Risc_Str_pack is

 -- Konvertierfunktionen in Textdarstellungen
 function str_RegAdr(RegAdr: tRegAdr) return string; -- Registeradressen
 function str_Opcode(Opcode: tOpcode) return string; -- Operationscode
 function str_Bed(Bed: tBed) return string;          -- Sprungbedingung
 function str_Instr(Instr: tInstr) return string;    -- Befehlswort
 function str(x: tOpTyp) return string;              -- Operationstyp
 function str(x: tErgebnis) return string;           -- Rechenwerkergebnis (Wert + Flags)

 -- Erzeugung eine Kopf und einer Datenzeile fuer den Prozessorstatus
 constant cProcStatTxt: string:=
  "|PC||    Instr     ||ORA|ORB|ORY1|OpCode|CtrlEX||  Ergebnis |CtrlRW|ORY2||r0 r1 r2 r3 r4 r5 r6 r7|";
 function str(pp: tPipeline; r: tDatMem(0 to 7)) return string;

 -- Ausgabe des Befehlsspeicherinhalts in Tabellenform (4 Spalten zu je 20 Zeichen)
 function str_array(x: tInstrMem) return tStringArray;

 -- Ausgabe des Datenspeicherinhaltes in Tabellenform (8 Spalten zu je 10 Zeichen)
 function str_array(x: tDatMem) return tStringArray;

end package;

package body Risc_Str_pack is
-------------------------------------------------------------------
-- Registeradersse => string
 constant RegSymTab: string:="r0 r1 r2 r3 r4 r5 r6 r7 ";
 function str_RegAdr(RegAdr: tRegAdr) return string is
 begin
  if is_x(RegAdr) then return "XX";
  else return str(RegAdr, RegSymTab);
  end if;
 end function;

-- Operationscode => string
 constant OpSymTab: string:="add inc sub dec neg sll srl sra rol ror not and or xor smov sadd ";
 function str_Opcode(Opcode: tOpcode) return string is
 begin
  if is_x(Opcode) then return "XX";
  else return str(Opcode, OpSymTab);
  end if;
 end function;

-- Sprungbedingung => string
 constant BedTab: string:="equ neq cy ncy neg geq ov nov ";
 function str_Bed(Bed: tBed) return string is
 begin
  if is_x(Bed) then return "XX";
  else return str(Bed, BedTab);
  end if;
 end function;

-- Befehlswort => string
 function str_Instr(Instr: tInstr) return string is
 begin
  if is_x(Instr) then 
   return "ungueltig";
  elsif Instr(cInstr_Init'range)=cInstr_Init then
   return "init " & str_RegAdr(Instr(tInstrRA'RANGE)) &
     ',' & str_hex(Instr(tInstrC8'RANGE)); 
  elsif Instr(cInstr_Op'range)=cInstr_Op then
   return str_Opcode(Instr(tInstrOp'RANGE)) & ' ' & str_RegAdr(Instr(tInstrRA'RANGE)) &
             ',' & str_RegAdr(Instr(tInstrRB'RANGE)); 
  elsif Instr(cInstr_st'range)=cInstr_st then
   return "st @(" & str_RegAdr(Instr(tInstrRB'RANGE)) & '+' & 
    str_hex(Instr(tInstrC2'RANGE)) & ")," & str_RegAdr(Instr(tInstrRA'RANGE));
  elsif Instr(cInstr_ldd'range)=cInstr_ldd then
   return "ldd " & str_RegAdr(Instr(tInstrRA'RANGE)) & ",@(" & str_RegAdr(Instr(tInstrRB'RANGE))
    & '+' & str(Instr(tInstrC2'RANGE)) & ')';
  elsif Instr(cInstr_ldc'range)=cInstr_ldc then
   return "ldc " & str_RegAdr(Instr(tInstrRA'RANGE)) & ",@(" & str_RegAdr(Instr(tInstrRB'RANGE))
    & '+' & str(Instr(tInstrC2'RANGE)) & ')';
  elsif Instr(cInstr_jmp'range)=cInstr_jmp then
   return "jmp " & str_RegAdr(Instr(tInstrRA'RANGE)) & ',' & str_RegAdr(Instr(tInstrRB'RANGE)); 
  elsif Instr(cInstr_jb'range)=cInstr_jb then
   return "jb_" & str_RegAdr(Instr(tInstrBed'RANGE)) & ' ' & str_RegAdr(Instr(tInstrRA'RANGE));    
  elsif Instr=cInstr_set_cy then return "set_cy";
  elsif Instr=cInstr_clr_cy then return "clr_cy";
  elsif Instr=cInstr_cpl_cy then return "cpl_cy";
  else return "nop";
  end if;
 end function;

-- Operationstyp => string
 function str(x: tOpTyp) return string is
 begin
  return tOpTyp'IMAGE(x);
 end function;

-- Ergebnis (Wert + Flags) => string
 function str(x: tErgebnis) return string is
 begin
  return str_hex(x.Wert) & ":c" & str(x.F.cy) & " s" &  str(x.F.s) & " z" & str(x.F.z);
 end function;

-------------------------------------------------------------------------
-- Konvertierfunktionen in Texttabellen
--
-- Erzeugung einer Datenzeile fuer den Prozessorstatus
-- |PC||    Instr     ||ORA|ORB|ORY1|OpCode|CtrlEX||  Ergebnis |CtrlRW|ORY2||r0 r1 r2 r3 r4 r5 r6 r7|

 function str(pp: tPipeline; r: tDatMem(0 to 7)) return string is
  variable s: tString;
 begin
  assign(s, "|" & str_hex(pp.PCount) & "||" & links(str_Instr(pp.Instr), 14) &
         "|| " & str_hex(pp.OReg_A) & "| " & str_hex(pp.OReg_B) &
        "| " & str_RegAdr(pp.ARegY1) & " | " & rechts(str_OpCode(pp.Opcode), 5) &  
         "|" & rechts(str(pp.Ctrl_EX), 5) & " ||" & rechts(str(pp.Ergebnis), 7) &
         "|" & rechts(str(pp.Ctrl_RW), 5) & " | " & str_RegAdr(pp.ARegY2) & " ||");
  for idx in r'low to r'high-1 loop
   append(s, str_hex(r(idx)) & ' ');
  end loop;
  append(s, str_hex(r(r'high)) & "|");
  return str(s);
 end function;

-- Ausgabe des Befehlsspeicherinhalts in Tabellenform (4 Spalten zu je 20 Zeichen)
 function str_array(x: tInstrMem) return tStringArray is
  variable StrArray: tStringArray(0 to (x'LENGTH-1)/4);
  variable adr: natural;
 begin
  for idx in StrArray'range loop
   for spalte in 0 to 3 loop
    adr:=idx + spalte*StrArray'length + x'low;
    if adr <= x'high then
     append(StrArray(idx), rechts(str(adr),3) & ':' & links(str_instr(x(adr)),16));
    else
     append(StrArray(idx), links("   :",20));
    end if;
   end loop;
  end loop;
  return StrArray;
 end function;

-- Ausgabe des Datenspeicherinhaltes in Tabellenform (8 Spalten zu je 10 Zeichen)
 function str_array(x: tDatMem) return tStringArray is
  variable StrArray: tStringArray(0 to (x'LENGTH-1)/8);
  variable adr: natural;
 begin
  for idx in StrArray'range loop
   for spalte in 0 to 7 loop
    adr:=idx + spalte*StrArray'length + x'low;
    if adr <= x'high then
     append(StrArray(idx), rechts(str(adr),3) & ':' & links(str_hex(x(adr)),7));
    else
     append(StrArray(idx), links("   :",11));
    end if;
   end loop;
  end loop;
  return StrArray;
 end function;
end package body;

