-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- 5.4 Entwurf eines RISC-Prozessors
--
-- Package für das Simulationsmodell eines RISC-Prozessors
--
-- Vereinbarung der Datentypen und Untertypen: tInstrAdr, tInstr, 
-- tRegAdr, tOpcode, tDaten, tInstrMem, tDatMem, tFlags, tErgebnis, 
-- tBed, tOpTyp und tPipeline
--
-- Vereinbarung der zugehoerigen Konstanten
--
-- Unterprogramme zur Modellierung der Funktion des Rechenwerks:
--  function ALU(Operand1, Operand2, Flags, Opcode) => tErgebnis;
--
-- Kontrolle der Sprungbedingung:
--  function Sprung (Bed; Flags) => boolean; -- True, wenn Sprung sicher
--  function SprungX(Bed; Flags) => boolean; -- True, wenn Sprung moeglich 
--
-- Funktionsbeschreibung fuer einen Berechnungsschritt
--  procedure PipeStep(Pipeline, InstrMem, RegSet, DatMem, KonstMem)
----------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
library Tuc;
use Tuc.Numeric_Sim.all;

package Risc_Pack is

-- Bitbreitendefinitionen aus Abschn. Befehlssatzentwurf
 constant cInstAdrBreite:natural:= 8;
 constant cInstrBreite:  natural:=12;
 constant cRegAdrBreite: natural:= 3;
 constant cOpcodeBreite: natural:= 4;
 constant cDatenBreite:  natural:= 8;

-- Typenvereinbarungen aus Abschn. Vearbeitungs-Pipeline 
 subtype tInstrAdr is std_logic_vector(cInstAdrBreite-1 downto 0);
 subtype tInstr    is std_logic_vector(cInstrBreite-1   downto 0);
 subtype tRegAdr   is std_logic_vector(cRegAdrBreite-1  downto 0);
 subtype tOpcode   is std_logic_vector(cOpcodeBreite-1  downto 0); 
 subtype tDaten    is std_logic_vector(cDatenBreite-1   downto 0);
 type tInstrMem is array (natural range <>) of tInstr;
 type tDatMem   is array (natural range <>) of tDaten;


 -- Bitvektorbereiche der Befehlsbestandteile aus Abschn. Befehlssatzentwurf
 type tInstrRA  is array  (2 downto 0) of std_logic; 
 type tInstrRB  is array  (5 downto 3) of std_logic; 
 type tInstrRY  is array  (2 downto 0) of std_logic; 
 type tInstrOp  is array  (9 downto 6) of std_logic; 
 type tInstrC8  is array (10 downto 3) of std_logic; 
 type tInstrC2  is array  (7 downto 6) of std_logic; 
 type tInstrBed is array  (5 downto 3) of std_logic; 

 -- konstante Befehlswortbestandteile aus Abschn. Befehlssatzentwurf
 constant cInstr_Init:std_logic_vector(11 downto 11):="1";
 constant cInstr_Op  :std_logic_vector(11 downto 10):="01";
 constant cInstr_st  :std_logic_vector(11 downto  8):="0010";
 constant cInstr_ldd :std_logic_vector(11 downto  8):="0011";
 constant cInstr_ldc :std_logic_vector(11 downto  8):="0001";
 constant cInstr_jmp :std_logic_vector(11 downto  6):="000011";
 constant cInstr_jb  :std_logic_vector(11 downto  6):="000010";

 -- parameterfrei Befehlsworte aus Abschn. Befehlssatzentwurf
 constant cInstr_nop    :tInstr:=x"000";
 constant cInstr_set_cy :tInstr:=x"001";
 constant cInstr_clr_cy :tInstr:=x"002";
 constant cInstr_cpl_cy :tInstr:=x"003";


-- Vereinbarungen aus Abschn. Rechenwerk
 constant opc_add: tOpcode:="0000";  
 constant opc_inc: tOpcode:="0001";  
 constant opc_sub: tOpcode:="0010"; 
 constant opc_dec: tOpcode:="0011"; 
 constant opc_neg: tOpcode:="0100"; 
 constant opc_sll: tOpcode:="0101"; 
 constant opc_srl: tOpcode:="0110"; 
 constant opc_sra: tOpcode:="0111"; 
 constant opc_rol: tOpcode:="1000";  
 constant opc_ror: tOpcode:="1001";
 constant opc_not: tOpcode:="1010"; 
 constant opc_and: tOpcode:="1011";  
 constant opc_or:  tOpcode:="1100"; 
 constant opc_xor: tOpcode:="1101";   
 constant sopc_mov:tOpcode:="1110";  
 constant sopc_add:tOpcode:="1111";

 -- Datentyp für die Flags
 type tFlags is record
  cy: std_logic; 
  z:  std_logic;                     
  s:  std_logic;                    
 end record;
-- tFlags, tErgebnis, tBed, tOpTyp, tPipeline
 -- Ergebnistyp der Rechenwerkfunktion
 type tErgebnis is record
  Wert: tDaten;
  F:   tFlags;
 end record; 

 function ALU(a,b:tDaten; F:tFlags; op:tOpcode) return tErgebnis;

-- Vereinbarungen aus Abschn. 5.3.4 Sprung-Pipeline
-- Sprung() wird TRUE, wenn der Sprung auszufuehren und 
-- SprungX() wird TRUE, wenn der Sprung moeglicher Weise 
-- auszufuehren ist
 subtype tBed is std_logic_vector(2 downto 0);
 function Sprung(Bed: tBed; F: tFlags) return boolean;
 function SprungX(Bed: tBed; F: tFlags) return boolean;

-- Vereinbarungen aus Abschn. 5.3.6 Gesamtentwurf
 type tOpTyp is (XX, non, wr, st, ldd, ldc,  setc, clrc, cplc);

 type tPipeline is record
  PCount          : tInstrAdr;
  Instr           : tInstr;
  OReg_A, OReg_B  : tDaten;
  Opcode          : tOpcode;
  ARegY1, ARegY2  : tRegAdr;
  Ergebnis        : tErgebnis; 
  Ctrl_EX, Ctrl_RW: tOpTyp;
 end record;

-- Funktionsbeschreibung fuer einen Berechnungsschritt
 procedure PipeStep(signal pp:inout tPipeline; 
         InstrMem:       tInstrMem; 
    signal RegSet: inout tDatMem(0 to 2**cRegAdrBreite-1);
    signal DatMem: inout tDatMem  ;
         KonstMem:       tDatMem  );  

 end RISC_pack;

 package body RISC_pack is

------------------------------------------------------------------
-- Ueberladen von "+" und "-" fuer std_logic_vector + Std_logic(_vector)
 function "+"(a, b: std_logic_vector) return std_logic_vector is
 begin
  return std_logic_vector(tUnsigned(a) + tUnsigned(b));
 end function;

 function "+"(a: std_logic_vector; b: std_logic) return std_logic_vector is
  constant cb: std_logic_vector(0 downto 0):= (0=>b);
 begin
  return std_logic_vector(tUnsigned(a) + tUnsigned(cb));
 end function;

 function "-"(a, b: std_logic_vector) return std_logic_vector is
 begin
  return std_logic_vector(tUnsigned(a) - tUnsigned(b));
 end function;

 function "-"(a: std_logic_vector; b: std_logic) return std_logic_vector is
  constant cb: std_logic_vector(0 downto 0):= (0=>b);
 begin
  return std_logic_vector(tUnsigned(a) - tUnsigned(cb));
 end function;

-- Funktion zur Beschreibung des Rechenwerks

 function ALU(a,b:tDaten; F:tFlags; op:tOpcode) return tErgebnis is
  variable y: tErgebnis;
  variable tmp: std_logic_vector(tDaten'LENGTH downto 0);
  variable neg, cy: std_logic;

 begin
  y.F:=F;
  case Op is
  -- Operationen ohne Statusänderung
   when opc_not|opc_and|opc_or|opc_xor|sopc_add|sopc_mov => 
    case op is

    -- bitweise Logikoperationen
     when opc_not => y.Wert :=   not b;
     when opc_and => y.Wert := a and b;
     when opc_or  => y.Wert := a  or b;
     when opc_xor => y.Wert := a xor b;

    --Spezialoperation fuer Laden, Speichern und Spruenge
     when sopc_add=> y.Wert := a + b;
     when sopc_mov=> y.Wert := a;
     when others => null;
    end case;

  -- Operation, die die Statusbits cy, s und z veraendern
   when others =>

   -- Addition und Subtraktion
    case Op is 
     when opc_add => tmp:= ('0'&a) + b + F.cy;
     when opc_inc => tmp:= ('0'&b) + '1';
     when opc_sub => tmp:= ('0'&a) - b - F.cy;
     when opc_dec => tmp:= ('0'&b) - '1';
     when opc_neg => tmp:= (not('0'&b)) +'1';

   -- Verschiebe- und Rotationsbefehle
     when opc_sll => tmp := b & '0';
     when opc_srl => tmp := b(0) & '0' & b(b'high downto 1);
     when opc_sra => tmp := b(0) & b(b'high) & b(b'high downto 1);
     when opc_rol => tmp := b & F.cy;
     when opc_ror => tmp := b(0) & F.cy & b(b'high downto 1);
     when others => tmp := (others =>'X');
    end case;

  -- Ergebnis- und Flag-Zuweisung
    y.Wert := tmp(tmp'high-1 downto tmp'low);
    y.F.cy:= tmp(tmp'high);
    y.F.s := tmp(tmp'high-1);
    if tmp(tmp'high-1 downto tmp'low)=x"00" then y.F.z := '1'; 
    elsif not is_x(tmp) then y.F.z := '0';
    else y.F.z := 'X';
    end if; 
  end case;
  return y;
 end function; 

------------------------------------------------------------------
-- Funktion zur Beschreibung der Sprungbedingung
-- Sprung() wird TRUE, wenn der Sprung auszufuehren und 
-- SprungX() wird TRUE, wenn der Sprung moeglicher Weise 
-- auszufuehren ist

 function Sprung(Bed: tBed; F: tFlags) return boolean is
  constant ov: BOOLEAN:= F.s /= F.cy;
 begin
  case Bed is
   when "000" => return F.z  = '1';  
   when "001" => return F.z  = '0';  
   when "010" => return F.cy = '1'; 
   when "011" => return F.cy = '0'; 
   when "100" => return F.s  = '1';  
   when "101" => return F.s  = '0';  
   when "110" => return ov;        
   when "111" => return not ov;  
   when others => return FALSE; 
  end case;
 end function; 

 function SprungX(Bed: tBed; F: tFlags) return boolean is
 begin
  case Bed is
   when "000"|"001" => return is_x(F.z); 
   when "010"|"011" => return is_X(F.cy); 
   when "100"|"101" => return is_x(F.s);  
   when "110"|"111" => return is_X(F.cy) or is_x(F.s);        
   when others => return TRUE;   
  end case;
 end function; 


------------------------------------------------------------------
-- Hilfsfunktionen und Prozeduren zum Lesen und Beschreiben
-- der Blockspeicher
 function read(Mem: tDatMem; adr: std_logic_vector) return tDaten is
  constant cXX: tDaten:= (others=>'X');
  variable idx: NATURAL;
 begin
  if is_x(adr) then return cXX;
  else
   idx:= int(tUnsigned(adr));
   if idx<Mem'low or idx>Mem'high then return cXX;
   else return Mem(idx);
   end if;
  end if;
 end function;

 function read(Mem: tInstrMem; adr: std_logic_vector) return tInstr is
  constant cXX: tDaten:= (others=>'X');
  variable idx: NATURAL;
 begin
  if is_x(adr) then return cXX;
  else
   idx:= int(tUnsigned(adr));
   if idx<Mem'low or idx>Mem'high then return cXX;
   else return Mem(idx);
   end if;
  end if;
 end function;

 procedure write(signal Mem: inout tDatMem; adr: std_logic_vector; x: tDaten) is
  constant cXX: tDaten:= (others=>'X');
  variable idx: NATURAL;
 begin
  if is_x(adr) then 
   for idx in Mem'range loop
    Mem(idx)<=cXX;
   end loop;
  else
   idx:= int(tUnsigned(adr));
   if idx>=Mem'low and idx<=Mem'high then 
    Mem(idx) <= x;
   end if;
  end if;
 end procedure;

-- Funktionsbeschreibung fuer einen Berechnungsschritt
 procedure PipeStep(signal pp:inout tPipeline; 
         InstrMem:       tInstrMem; 
    signal RegSet: inout tDatMem(0 to 2**cRegAdrBreite-1);
    signal DatMem: inout tDatMem  ;
         KonstMem:       tDatMem  ) is
 begin  
 -- Registerzuweisungen fuer Verarbeitungsoperationen
  pp.PCount<=pp.PCount+'1';
  pp.Instr<=read(InstrMem, pp.PCount);
  pp.OReg_A<=read(RegSet, pp.Instr(tInstrRA'RANGE)); 
  pp.OReg_B<=read(RegSet, pp.Instr(tInstrRB'RANGE)); 
  pp.OpCode<=pp.Instr(tInstrOp'RANGE);
  pp.ARegY1<=pp.Instr(tInstrRY'RANGE);
  pp.Ctrl_EX<=wr;
  -- pp.Ergebnis<=ALU(...); => Beschreibung EX-Phase
  pp.ARegY2<=pp.ARegY1; 
  pp.Ctrl_RW<=pp.Ctrl_EX;

 -- Veraenderungen der Registerzuweisungen in der OF-Phase
  if pp.Instr(cInstr_Init'range)=cInstr_Init then
   pp.OReg_A<=pp.Instr(tInstrC8'RANGE);  
   pp.OpCode<=sopc_mov; 
   pp.Ctrl_EX<=wr;
  elsif pp.Instr(cInstr_Op'range)=cInstr_Op then
   null;
  elsif pp.Instr(cInstr_st'range)=cInstr_st then
   pp.OReg_A<="000000" & pp.Instr(tInstrC2'RANGE);
   pp.OpCode<=sopc_add;
   pp.Ctrl_EX<=st;
  elsif pp.Instr(cInstr_ldd'range)=cInstr_ldd then
   pp.OReg_A<="000000" & pp.Instr(tInstrC2'RANGE);
   pp.OpCode<=sopc_add;
   pp.Ctrl_EX<=ldd;
  elsif pp.Instr(cInstr_ldc'range)=cInstr_ldc then
   pp.OReg_A<="000000" & pp.Instr(tInstrC2'RANGE);
   pp.OpCode<=sopc_add;
   pp.Ctrl_EX<=ldc;
  elsif pp.Instr(cInstr_jmp'range)=cInstr_jmp then
   pp.PCount <= read(RegSet, pp.Instr(tInstrRB'RANGE));
   pp.OReg_A<=pp.PCount+'1'; 
   pp.OpCode<=sopc_mov;  
   pp.Ctrl_EX<=wr;
  elsif pp.Instr(cInstr_jb'range)=cInstr_jb then
   if SprungX(pp.Instr(tInstrBed'RANGE), pp.Ergebnis.F) then
    pp.PCount <= (others=>'X'); 
   elsif Sprung(pp.Instr(tInstrBed'RANGE), pp.Ergebnis.F) then
    pp.PCount <= read(RegSet, pp.Instr(tInstrRB'RANGE));
   end if;
  elsif pp.Instr=cInstr_set_cy then pp.Ctrl_EX<=setc;
  elsif pp.Instr=cInstr_clr_cy then pp.Ctrl_EX<=clrc;
  elsif pp.Instr=cInstr_cpl_cy then pp.Ctrl_EX<=cplc;
  elsif not is_x(pp.Instr)     then pp.Ctrl_EX<=non;
  else 
  -- bei einem ungueltigen Befehlswort werden alle Registerinhalte,
  -- die aus dem Befehlswort gebildet werden, ungueltig
   pp.OReg_A<=(others=>'X'); pp.OReg_B<=(others=>'X');
   pp.Opcode<=(others=>'X'); pp.ARegY1<=(others=>'X');
   pp.Ctrl_EX<=XX; 
  end if;

 -- Ergebniszuweisungen in der EX-Phase
 case pp.Ctrl_EX is
  when setc => pp.Ergebnis.F.cy<='1';
  when clrc => pp.Ergebnis.F.cy<='0';
  when cplc => pp.Ergebnis.F.cy<=not pp.Ergebnis.F.cy;
  when non  => null;
  when XX   => pp.Ergebnis <= (Wert => (others=>'X'), 
       F => (cy=>'X', s=>'X', z=>'X')); 
  when others => 
   pp.Ergebnis<=ALU(pp.OReg_A, pp.OReg_B, pp.Ergebnis.F, pp.Opcode); 
 end case;

 -- Schreiboperationen in der RW-Phase
 case pp.Ctrl_RW is
  when wr  => write(RegSet, pp.ARegY2, pp.Ergebnis.Wert);
  when st  => write(DatMem, pp.Ergebnis.Wert, read(RegSet, pp.ARegY2));
  when ldd => write(RegSet, pp.ARegY2, read(DatMem, pp.Ergebnis.Wert));
  when ldc => write(RegSet, pp.ARegY2, read(KonstMem, pp.Ergebnis.Wert));
  when XX =>
   write(RegSet, pp.ARegY2, (others=>'X'));
   write(DatMem, pp.Ergebnis.Wert, (others=>'X'));
  when others => null;
 end case;
end procedure;

end package body;


