----------------------------------------------------------------
-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- Bibliothek Tuc
----------------------------------------------------------------
-- Package Tuc.Ausgabe
--
-- Grundfunktionen für die Textausgabe
--
-- Textobjekte: 
--  tString       Textzeile variabler Länge
--  tStringArray  Feld von Zeichenketten variabler Laenge (mehrere Textzeilen)
--
-- Bearbeitungsmethoden für Zeichenketten variabler Länge
--  * assign(ts, s))       -- einer Textzeile eine Zeichenkette s zuweisen
--  * clear(ts)            -- Textzeile leeren
--  * append(ts, s)        -- eine Zeichenkette s an eine Textzeile anhängen s anhaengen
--
-- Ausgabeanweisungen
--  * write(f, ts)         -- mit einer Zeile aus Text-Datei f initialisieren
--  * write(ts)         -- mit einer Zeile aus Text-Datei f initialisieren
--  * write(f, tsa)         -- mit einer Zeile aus Text-Datei f initialisieren
--  * write(tsa)         -- mit einer Zeile aus Text-Datei f initialisieren
-- (f -- Textdatei (file of text); s -- Zeichenkette (string);
-- ts -- Textzeile (tString); tsa -- Feld von Textzeilen (tStringArray))
--
-- Konvertierung von Datenobjekten in Ausgabetexte
--
-- str(x [,fmt]) -- Konvertieren von x in eine Textdarstellung
--   x:   BIT, BOOLEAN, STD_ULOGIC, BIT_VECTOR, STD_LOGIC_VECTOR, 
--      INTEGER, REAL, TIME; tString
--   fmt: Formatbuchstabe: 't' -- Typ, 'b' -- binaer, 'o' -- oktal, 'h' hexadezimal,
--   'd' -- dezimal, 's' -- dezimal mit vorzeichenbehaftete, ...)
--
-- str(x, SymbolTab) -- Darstellen der Werte von x als Tabellensymbole 
--   (x: INTEGER, BIT_VECTOR, STD_LOGIC_VECTOR)
-- Die Symboltabelle ist eine Zeichenkette, in der die Symbole in der
-- die Symbole in aufsteigender Reihenfolge gefolgt von einem Leerzeichen 
-- stehen, z.B. "load store add sub mult div ". 
--
-- Verlängerung einer Zeichenkette s auf die Länge n
--
-- rechts(s, n) Rechtsausrichtung (links Leerzeichen anhängen)
-- links(s, n)  Linksausrichtung (rechts Leerzeichen anhängen)
--
-- 15.05.2010, G. Kemnitz
----------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use std.textio.TEXT;
use std.textio.WRITE;
use std.textio.OUTPUT;
--use IEEE.MATH_REAL;
use IEEE.NUMERIC_BIT.all;
 
package Ausgabe is

--------------------------------------------------------------------
-- Datenobjekt und Prozeduren zum Zusammensetzen von Textzeilen
 constant zeilenlaenge: POSITIVE:=100;
 type tString is record 
  str: STRING( 1 to zeilenlaenge);
  ptr: POSITIVE;
 end record;

-- mehrzeiliges Textobjekt
 type tStringArray is array (NATURAL range <>) of tString;

 function str(s: tString; fmt: CHARACTER:='s') return STRING;
 procedure append(x: inout tString; s: STRING);
 procedure clear(x: inout tString);
 procedure assign(x: inout tString; s: string);

-- Ausgabe auf die Standardausgabe
 procedure write(s: STRING);
 procedure write(file f: text; ts: tString);
 procedure write(ts: tString);
 procedure write(file f: text; tsa: tStringArray);
 procedure write(tsa: tStringArray);

---------- Konvertierfunktionen in Zeichenketten ------------
-- <Zeichenkette> := str(<Datenobjekt>[,<Anz. Nachkommastellen>][,Zahlenbasis])
-- der kleinste Integer-Wert wird bei der Simulation auch als Pseudo-Wert fuer 
-- verwendet
 constant int_invalid: INTEGER:=INTEGER'low; 

 function str(x: CHARACTER; fmt:CHARACTER:='c') return STRING;
 function str(x: BIT; fmt:CHARACTER:='b') return STRING;
 function str(x: STD_ULOGIC; fmt:CHARACTER:='b') return STRING;
 function str(x: BOOLEAN; fmt:CHARACTER:='b') return STRING;

 function str(x: BIT_VECTOR; fmt:CHARACTER:='b') return STRING;
 function str(x: STD_ULOGIC_VECTOR; fmt:CHARACTER:='b') return STRING;
 function str(x: STD_LOGIC_VECTOR; fmt:CHARACTER:='b') return STRING;
 function str(x: INTEGER; fmt: CHARACTER:='d') return STRING;
 function str(x: REAL; fmt: CHARACTER:='2') return STRING;
 function str(x: REAL; m: POSITIVE) return STRING;
 function str(t: time; fmt: CHARACTER:='2') return STRING;

----------------------------------------------------------------
-- spezielle Ausgabefunktion fuer das Projekt .. (RISC-Prozesser)
-- (Darstellung ohne Kennzeichnung der Zahlenbasis)
 function str_hex(x: std_logic_vector) return string;

-- mit Uebersetzungstabelle in symbolische Werte
-- Die Symboltabelle ist ein Zeichenkette der Symbole
-- mit einem Leerzeichen hinter jedem Symbol; z.B. "add sub "
-- Zahlewert    Symbol
--   0           add
--   1           sub
--  ...          ...
 function str(n: NATURAL; SymbolTab: STRING) return STRING;
 function str(x: BIT_VECTOR; SymbolTab: STRING) return STRING;
 function str(x: STD_ULOGIC_VECTOR; SymbolTab: STRING) return STRING;
 function str(x: STD_LOGIC_VECTOR; SymbolTab: STRING) return STRING;


-- Funktion zum Ausrichten einer Zeichenkette
 function rechts(x: STRING; n: POSITIVE) return STRING;
 function links(x: STRING; n: POSITIVE) return STRING;

 
end package;
--=============================================================
package body Ausgabe is


-- Prozeduren zum Zusammensetzen von Textzeilen

 procedure assign(x: inout tString; s: STRING) is
 begin
  x.ptr := 1;
  append(x, s);
 end procedure;

 procedure clear(x: inout tString) is
 begin
  x.ptr := 1;
 end procedure;

 procedure append(x: inout tString; s: STRING) is
 begin
  for idx in s'RANGE loop
  if x.ptr>x.str'HIGH then exit;
   else
    x.str(x.ptr):=s(idx);
    x.ptr := x.ptr+1;
   end if;
  end loop;
 end procedure;

-----------------------------------------------------------------
 
 function str(s: tString; fmt: CHARACTER:='s') return STRING is
 begin
  if fmt='t' then return "tString";
  else return s.str(1 to s.ptr-1);
  end if;
 end function;

-----------------------------------------------------------------
-- Standardausgabe

 procedure WRITE(s: STRING) is
 begin
  WRITE(OUTPUT, s);
 end procedure;

procedure write(ts: tString) is
 begin
  if ts.ptr<=ts.str'LENGTH then write(ts.str(1 to ts.ptr));
  else write(ts.str);
  end if;
 end procedure;

 procedure write(file f: text; ts: tString) is
 begin
  WRITE(f, ts.str(1 to ts.ptr));
 end procedure;

 procedure write(file f: text; tsa: tStringArray) is
 begin
  for idx in tsa'range loop
   write(f, tsa(idx));
  end loop;
 end procedure;

 procedure write(tsa: tStringArray) is
 begin
  for idx in tsa'range loop
   write(tsa(idx));
  end loop;
 end procedure;

-----------------------------------------------------------------

--======================================================
-- nicht exportierte Hilfsfunktionen
-- Konvertieren eines Bitvektors in eine Hex-Zifferfolge

 function char_hex(x: BIT_VECTOR(3 downto 0)) return CHARACTER is
 begin
  case x is
   when "0000" => return '0';
   when "0001" => return '1';
   when "0010" => return '2';
   when "0011" => return '3';
   when "0100" => return '4';
   when "0101" => return '5';
   when "0110" => return '6';
   when "0111" => return '7';
   when "1000" => return '8';
   when "1001" => return '9';
   when "1010" => return 'a';
   when "1011" => return 'b';
   when "1100" => return 'c';
   when "1101" => return 'd';
   when "1110" => return 'e';
   when "1111" => return 'f';
  end case;
 end function;

function str_hex(x: BIT_VECTOR) return STRING is
 constant xx: BIT_VECTOR(x'LENGTH+2 downto 0):="000"&x;
 variable y: STRING(1 to xx'LENGTH/4);
begin
 --write("Test xx'length=:"&str(xx'length));
 --write("Test y'length=:"&str(y'length));
 for idx in y'range loop
  --write(str(4*(y'LENGTH-idx)+3)&':'& str(4*(y'LENGTH-idx)));
  y(idx):= char_hex(xx(4*(y'LENGTH-idx)+3 downto 4*(y'LENGTH-idx)));
 end loop;
 return y&'h';
end function;

-- spezielle Ausgabefunktion fuer Projekt .. (RISC-Prozessor)
 function str_hex(x: std_logic_vector) return string is
  constant tmp: std_logic_vector(x'LENGTH+2 downto 0):="000"&x;
  variable y: STRING(1 to tmp'LENGTH/4);
 begin
  if is_x(x) then
   y:=(others=>'X');
  else
   for idx in y'range loop
    y(idx):= char_hex(to_bitvector(tmp(4*(y'LENGTH-idx)+3 downto 4*(y'LENGTH-idx))));
   end loop;
  end if;
  return y;
 end function;


-- Konvertieren eines Bitvektors in eine Oktal-Ziffernfolge

 function char_oct(x: BIT_VECTOR(2 downto 0)) return CHARACTER is
 begin
  case x is
   when "000" => return '0';
   when "001" => return '1';
   when "010" => return '2';
   when "011" => return '3';
   when "100" => return '4';
   when "101" => return '5';
   when "110" => return '6';
   when "111" => return '7';
  end case;
 end function;

function str_oct(x: BIT_VECTOR) return STRING is
 constant xx: BIT_VECTOR(x'LENGTH+1 downto 0):="00"&x;
 variable y: STRING(1 to xx'LENGTH/3);
begin
 --write("Test x'length=:"&str(xx'length));
 --write("Test y'length=:"&str(y'length));
 for idx in y'range loop
  --write(str(3*(y'LENGTH-idx)+2)&':'& str(3*(y'LENGTH-idx)));
  y(idx):= char_oct(xx(3*(y'LENGTH-idx)+2 downto 3*(y'LENGTH-idx)));
 end loop;
 return y&'o';
end function;

-- Konvertieren eines Bitvektors in eine binäre Ziffernfolge
function str_bin(bv: bit_vector) return STRING is
 variable y: STRING(1 to bv'length);
 variable idx: positive:=1;
begin
 for id in bv'RANGE loop
  if bv(id)='1' then y(idx):= '1';
  else y(idx):= '0';
  end if;
  idx:=idx+1;
 end loop;
 return y&'b';
end function;


-- Konvertierung einer Gleitkommazahl in eine Festkommadarstellung
 function str(x: REAL; m: POSITIVE) return STRING is
  constant z: POSITIVE := 10**m;
  constant i: INTEGER := INTEGER(REAL(z)*x);
  constant snk: STRING(1 to m+1):= str(z+ abs(i rem z));
  variable a:INTEGER;
 begin
  if abs(x)>REAL(INTEGER'HIGH) then return REAL'IMAGE(x);
  elsif i<0 and i/z=0 then 
   return '-'&str(i/z) & "."&snk(2 to snk'LENGTH);
  else
   return str(i/z) & "."&snk(2 to snk'LENGTH);
  end if;
 end function;

--========================================================================
-- exportierte Funktionen

---- character
function str(x: CHARACTER; fmt:CHARACTER:='c') return STRING is
 constant s: STRING(1 to 1):=(1=>x);
begin
 if fmt='t' then return "character";
 else return s;
 end if;
end function;

---- bit
 function str(x: BIT; fmt:CHARACTER:='b') return STRING is
 begin
  if fmt='t' then return "bit";
  elsif x='1' then return "1";
  else return "0";
  end if;
 end function;

---- std_ulogic
 constant SLV_Tab: STRING:="UX01ZWLH-";
 function str(x: STD_ULOGIC; fmt:CHARACTER:='b') return STRING is
 begin
  if fmt='t' then return "std_ulogic";
  else return str(SLV_Tab(std_ulogic'pos(x)+1));
  end if;
 end function;

---- boolean
 function str(x: BOOLEAN; fmt:CHARACTER:='b') return STRING is
 begin
  if fmt='t' then return "BOOLEAN";
  else return BOOLEAN'IMAGE(x);
  end if;
 end function;

---- bit_vector
 function str(x: BIT_VECTOR; fmt:CHARACTER:='b') return STRING is
 begin
  case fmt is
   when 't' => return "bit_vector";
   when 'o' => return str_oct(x);
   when 'h' => return str_hex(x);
   when 'd' => return str(TO_INTEGER(UNSIGNED(x)));
   when 's' => return str(TO_INTEGER(SIGNED(x)));
   when others => return str_bin(x);
  end case;
 end function;

---- std_ulogic_vector
 function str(x: STD_ULOGIC_VECTOR; fmt:CHARACTER:='b') return STRING is
 begin 
  if fmt='t' then return "std_ulogic_vector";
  else return str(STD_LOGIC_VECTOR(x), fmt);
  end if;
 end function;

---- std_logic_vector
 function str(x: STD_LOGIC_VECTOR; fmt:CHARACTER:='b') return STRING is
  variable s: STRING(1 to x'LENGTH);
  variable idx: POSITIVE:=1;
 begin
  if fmt='t' then return "std_logic_vector";
  elsif not IS_X(x) then
   case fmt is 
    when 'h' => return str_hex(TO_BITVECTOR(x));
    when 'o' => return str_oct(TO_BITVECTOR(x));
    when 'd' => return str(TO_INTEGER(UNSIGNED(TO_BITVECTOR(x))));
    when 's' => return str(TO_INTEGER(SIGNED(TO_BITVECTOR(x))));
    when others => null;
   end case;
  end if;
  for id in x'RANGE loop
   s(idx):=SLV_Tab(std_ulogic'pos(x(id))+1);
   idx:=idx+1;
  end loop;
  return s;
 end function;

---- integer
 function str(x: INTEGER; fmt: CHARACTER:='d') return STRING is
 begin
  case fmt is
   when 't' => return "integer";
   when 'h' => return str_hex(BIT_VECTOR(TO_SIGNED(x, 32)));
   when 'o' => return str_oct(BIT_VECTOR(TO_SIGNED(x, 32)));
   when 'b' => return str(BIT_VECTOR(TO_SIGNED(x, 32)));
   when others => 
    if x=int_invalid then return "invalid";
    else return INTEGER'IMAGE(x);
    end if;
  end case;
end function;

---- real

 function str(x: REAL; fmt: CHARACTER:='2') return STRING is
 begin
  case fmt is
   when 't' => return "real";
   when '1'|'2'|'3'|'4' => return str(x, character'POS(fmt)-CHARACTER'POS('0'));
   when others => return REAL'IMAGE(x);
  end case;
 end function;

---- time

 function str(t: TIME; fmt: CHARACTER:='2') return STRING is
  variable m: NATURAL;
 begin
  case fmt is
   when 't' => return "time";
   when '1'|'2'|'3'|'4' => m:=character'POS(fmt)-CHARACTER'POS('0');
   when others => m:=2;
  end case;
  if    abs(t)< 1 ps then return str(t/fs) & " fs";
  elsif abs(t)< 1 ns  then return str(1.0E-3*REAL(t/fs), m) & " ps";
  elsif abs(t)< 1 us  then return str(1.0E-6*REAL(t/fs), m) & " ns";
  elsif abs(t)< 1 ms  then return str(1.0E-6*REAL(t/ps), m) & " us";
  elsif abs(t)< 1 sec then return str(1.0E-6*REAL(t/ns), m) & " ms";
  elsif abs(t)< 1 min then return str(1.0E-6*REAL(t/us), m) & " sec";
  else return str(1.0E-6*REAL(t/60 us),  m) & " min";
  end if;
 end function;

-- NATURAL mit Uebersetzungstabelle in symbolische Werte
function str(n: NATURAL; SymbolTab: STRING) return STRING is
 variable a, m: NATURAL;
begin 
 for idx in SymbolTab'RANGE loop
  if SymbolTab(idx)=' ' then m:=m+1;
   if m=n then a:=idx;
   elsif m=n+1 then return SymbolTab(a+1 to idx-1);
 end if; end if; end loop;
 return "Sym("&str(n)&')';
end function;

-- Bitvektor mit Uebersetzungstabelle in symbolische Werte
function str(x: BIT_VECTOR; SymbolTab: STRING) return STRING is
begin
 return str(TO_INTEGER(UNSIGNED(x)), SymbolTab);
end function;

function str(x: STD_LOGIC_VECTOR; SymbolTab: STRING) return STRING is
begin
 if IS_X(x) then return "Sym("&str(x)&')';
 else return str(TO_BITVECTOR(x), SymbolTab);
 end if;
end function;

function str(x: STD_ULOGIC_VECTOR; SymbolTab: STRING) return STRING is
begin
 return str(STD_LOGIC_VECTOR(x), SymbolTab);
end function;

-- ====================================================================
-- Funktion zum Ausrichten einer Zeichenkette
 function rechts(x: STRING; n: POSITIVE) return STRING is
  variable s: STRING(1 to n):=(others=>' ');
 begin
  if x'LENGTH>=n then return x;
  else return s(1 to n-x'LENGTH)&x;
  end if;
 end function;

 function links(x: STRING; n: POSITIVE) return STRING is
  variable s: STRING(1 to n):=(others=>' ');
 begin
  if x'LENGTH>=n then return x;
  else return x&s(1 to n-x'LENGTH);
  end if;
 end function;


end package body;

