----------------------------------------------------------------
-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- Bibliothek Tuc
----------------------------------------------------------------
-- Package Tuc.Eingabe
--
-- tPString: Parster-Datenobjekt; Bearbeitungsmethoden
--  * read(f, pstr)            -- mit einer Zeile aus Text-Datei f initialisieren
--  * assign(pstr, s))         -- mit einer Zeichenkette s initialisieren
--  * clear(pstr)              -- leeren
--  * append(pstr, s)          -- Zeichenkette s anhaengen
--  * get_char(pstr, pos, CharTab)  -- Zeichen aus CharTab vom Band nehmen
--  * get_sym(pstr, pos, SymbolTab) -- Zeichen aus Symboltabelle vom Band nehmen
--  * get(pstr, w [,min, max]) -- Datenobjekte w vom Band nehmen
--  * read(promt, w)           -- Datenobjekt w von der Standardeingabe lesen
--  * read		       -- Warte auf Betaetigung der Eingabetaste
--
-- (pstr -- Parser-Objekt; w -- BIT, BOOLEAN, STD_ULOGIC, BIT_VECTOR, 
-- STD_LOGIC_VECTOR, INTEGER, REAL, TIME; promt -- Ausgabetext)
--
-- 15.01.2010, G. Kemnitz
----------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use STD.TEXTIO.all;
library tuc;
use Tuc.Ausgabe.all;
use Tuc.StopSim_pack.all;


package Eingabe is

 constant debug:boolean:=False;

 type tLesestatus is (X, ok, trunc, err); 

 type tPString is record
  str: string(1 to zeilenlaenge); -- Zeichenkette
  pta, pte: POSITIVE;             -- Anfangs- und Endzeiger Verarbeitungsposition
  Status: tLesestatus;            -- Status der letzen Operation
  err_pos: POSITIVE;		  -- Fehlerposition
  err_msg: tString;		  -- Fehlermeldung
 end record;

-- Unterprogramme zum Beschreiben und Leeren eines Parser-Objekts
 procedure read(file f: TEXT; pstr: out tPString);
 procedure append(pstr: inout tPString; s: string);
 procedure clear(pstr: inout tPString);
 procedure assign(pstr: inout tPString; s: string);
 function str(pstr: tPString) return string;

-- Abräumen eines Zeichens aus einer Zeichentabelle, Rueckgabe der Position
-- Die Symbole in der Symboltabelle werden mit dem Trennzeichen TZ getrennt 
 procedure get_char(pstr: inout tPString; pos: out NATURAL; CharTab: string);
 procedure get_sym(pstr: inout tPString; pos: out NATURAL; SymbolTab: string; TZ: CHARACTER:=' ');

 -- Abräumen aller Zeichen, die in CharTab stehen (Leerzeichen entfernen) 
 procedure skip(pstr: inout tPString; CharTab: string:=" ");

-- 
 procedure get(pstr: inout tPString; w: out INTEGER;
           wmin: INTEGER:=INTEGER'low+1; wmax: INTEGER:= INTEGER'high);
 procedure get(pstr: inout tPString; w: out REAL; 
           wmin: REAL:=REAL'low/10.0; wmax: REAL:= REAL'high/10.0);
 procedure get(pstr: inout tPString; w: out TIME; 
           wmin: TIME:=0 fs; wmax: TIME:= 1 sec);
 procedure get(pstr: inout tPString; w: out BIT);
 procedure get(pstr: inout tPString; w: out BOOLEAN);
 procedure get(pstr: inout tPString; w: out STD_ULOGIC);
 procedure get(pstr: inout tPString; w: out BIT_VECTOR);
 procedure get(pstr: inout tPString; w: out STD_LOGIC_VECTOR);

-- Einlesen von der Standardeingabe
 procedure CheckExit(pstr: inout tPString);
 procedure read(promt: string; w: out tPString);
 procedure read(promt: string; w: out INTEGER; 
           wmin: INTEGER:=INTEGER'low+1; wmax: INTEGER:= INTEGER'high);
 procedure read(promt: string; w: out REAL; 
           wmin: REAL:=REAL'low/10.0; wmax: REAL:= REAL'high/10.0);
 --procedure read(promt: string; w: out TIME; 
 --          wmin: TIME:=0 fs; wmax: TIME:= TIME'high/10);
 procedure read(promt: string; w: out BIT);
 procedure read(promt: string; w: out BOOLEAN);
 procedure read(promt: string; w: out STD_ULOGIC);
 procedure read(promt: string; w: out BIT_VECTOR);
 procedure read(promt: string; w: out STD_LOGIC_VECTOR);
 procedure read;


end package;




package body Eingabe is

-- Anhaengen einer Zeichenkette an ein PString-Objekt
 procedure append(pstr: inout tPString; s: STRING) is
 begin
  if s'LENGTH < pstr.str'LENGTH - pstr.pte +1 then
   pstr.str(pstr.pte to pstr.pte+s'LENGTH-1) := s;
   pstr.pte := pstr.pte + s'LENGTH;
   pstr.Status:=ok;
  elsif pstr.pte < pstr.str'LENGTH then
   pstr.str(pstr.pte to pstr.str'LENGTH) := s(s'LOW to pstr.str'LENGTH-pstr.pte+1);
   pstr.pte := pstr.str'LENGTH;
   pstr.Status:=trunc;
  end if;
 end procedure;

-- leeren eines PString-Objekts
 procedure clear(pstr: inout tPString) is
 begin
  pstr.pta:=1;
  pstr.pte:=1;
  pstr.Status:=ok;
 end procedure;

-- leeren eines PString-Objekts und eine neue Zeichenkette anhängen
 procedure assign(pstr: inout tPString; s: STRING) is
 begin
  clear(pstr);
  append(pstr, s);
 end procedure;

 function str(pstr: tPString) return STRING is
 begin
  case pstr.Status is
   when ok => 
    if pstr.pte>pstr.pta then
     return '['&str(pstr.pta)&'|'& pstr.str(pstr.pta to pstr.pte-1) &'|'& str(pstr.pte)&']';
    else return "[EOL]";
    end if;
     
   when err => return pstr.str(pstr.pta to pstr.err_pos-1) &"..=>"&pstr.str(pstr.err_pos to pstr.pte-1) & "  ["& str(pstr.err_msg)&"]";
   when trunc => return pstr.str(pstr.pta to pstr.pte-1) & " [trunc]";
   when others => return pstr.str(pstr.pta to pstr.pte-1) & " [X]";
  end case;
 end function;

-- Abraeumen eines Zeichens aus einer Zeichentabelle
procedure get_char(pstr: inout tPString; pos: out natural; CharTab: STRING) is
 variable c: CHARACTER:=pstr.str(pstr.pta);
begin
 --write("pstr.pta: "&str(pstr.pta) & "  pstr.str(pstr.pta): " & str(pstr.str(pstr.pta+1)));
 --pos:=0;
 if pstr.Status=ok and pstr.pta< pstr.pte then 

  for idx in CharTab'RANGE loop
   --write("c : CharTab(idx) " &str(c)& ':' & str(CharTab(idx)));
   if CharTab(idx)=c then 
    pos:=idx-1; 
    if pstr.pta<pstr.pte then
     pstr.pta:=pstr.pta+1;
     --write("pta:pte "&str(pstr.pta)&':'&str(pstr.pte));
    end if;
    --write("pos: "&str(idx-1));
    return;
   end if;
  end loop;
  pstr.Status:=err;
  pstr.err_pos:=pstr.pta;
  assign(pstr.err_msg, "expected characters: "&CharTab);
 end if;
  --write("TT4: pstr:"&str(pstr));
end procedure;


-- Abraeumen eines Symbols aus einer Symboltabelle

 procedure get_sym(pstr: inout tPString; pos: out NATURAL; SymbolTab: STRING; TZ: CHARACTER:=' ') is
  constant pta: POSITIVE:=pstr.pta;
  variable ct: NATURAL;
  variable gleich: BOOLEAN:=TRUE;
  variable line_end: BOOLEAN:=False;
 begin
  if pstr.Status = ok and pstr.pta< pstr.pte then -- pstr.Status=ok and 
   for idx in SymbolTab'RANGE loop 
    --write("SymbolTab(idx):pstr.str(pstr.pta) "&SymbolTab(idx)&':'&pstr.str(pstr.pta));
    if SymbolTab(idx)=TZ then
     if gleich then
      pos:=ct; 
      return;
     else
      pstr.pta:= pta;
      gleich:=TRUE;
      ct:=ct+1;
     end if;
    else 
     if pstr.pta>=pstr.pte or SymbolTab(idx)/=pstr.str(pstr.pta) then gleich:=FALSE; end if;
     pstr.pta:=pstr.pta+1;
    end if;
   end loop;
 end if;
 pstr.pta:=pta;
 pstr.Status:=err;
 pstr.err_pos:=pta;
 assign(pstr.err_msg, "expected symbols: "&SymbolTab);
end procedure;

 -- Abraeumen aller Zeichen, die in CharTab stehen (Leerzeichen entfernen),  
 procedure skip(pstr: inout tPString; CharTab: STRING:=" ") is
  variable match: boolean;
 begin
  while pstr.Status = ok and pstr.pta< pstr.pte-1 loop
   for idx in CharTab'range loop
    if CharTab(idx)=pstr.str(pstr.pta) then
     match:=True;
     pstr.pta:=pstr.pta+1;
     exit;
    end if;
    match:=False;
   end loop;
   exit when not match;
  end loop;
 end procedure;


-- Abraeumen einer Ziffernfolge (Hilfsprozedur)
 procedure getzw(pstr: inout tPString; w: out real) is
  constant ptr: POSITIVE:=pstr.pta;
  variable digit: natural;
  variable val: REAL;
 begin
  if pstr.Status=ok and pstr.pta< pstr.pte then
   get_char(pstr, digit, "0123456789");
   if pstr.Status =ok then
    val:=real(digit);
    --write("T1 pstr="&str(pstr)&" digit="&str(digit)&" val="&str(val));
    while pstr.Status=ok and pstr.pta< pstr.pte loop
     get_char(pstr, digit, "0123456789");
     --write("T2 pstr="&str(pstr)&" digit="&str(digit));
     exit when pstr.Status/=ok;
     val:=10.0*val+REAL(digit);
     --write("Tn pstr="&str(pstr)&" digit="&str(digit)&" val="&str(val));
    end loop;
    pstr.STATUS:=ok;
    w:=val;
   end if;
  end if;
 end procedure;
 

-- Abraeumen eines Integers
 procedure get(pstr: inout tPString; w: out INTEGER; 
           wmin: INTEGER:=INTEGER'low+1; wmax: INTEGER:= INTEGER'high) is
  constant ptr: POSITIVE:=pstr.pta;
  variable sign, digit: natural;
  variable val: REAL;
 begin
  if pstr.Status=ok and pstr.pta< pstr.pte then
   get_char(pstr, sign, "+-");
   if not (pstr.Status=ok) then 
    sign:=0;
    pstr.Status:= ok;
   end if;
   getzw(pstr, val);
   if sign=1 then val:=-val; end if;
    if pstr.Status =ok and val>=REAL(wmin) and val <=REAL(wmax) then 
     w:= integer(val);
     return;
    else
     pstr.Status:=err;
    end if;
    pstr.pta:=ptr;
    pstr.err_pos:=ptr;
    assign(pstr.err_msg, "integer range("&str(wmin)&" to "&str(wmax)&") expected");
   end if;
  end procedure;

-- Abraeumen einer Gleitkommazahl
procedure get(pstr: inout tPString; w: out REAL; 
           wmin: REAL:=REAL'low/10.0; wmax: REAL:= REAL'high/10.0) is
  variable ptr: POSITIVE:=pstr.pta;
  variable sign, pos: natural;
  variable val, val_nk: REAL:=0.0;
 begin
  if pstr.Status=ok and pstr.pta< pstr.pte then
   get_char(pstr, sign, "+-");
   if not (pstr.Status=ok) then 
    sign:=0;
    pstr.Status:= ok;
   end if;
   --write("Vorzeichen:" & str(sign));
   getzw(pstr, val);
   if pstr.Status=ok then
    get_char(pstr, pos, ".");
    if pstr.Status= ok then
     ptr:=pstr.pta;
     getzw(pstr, val_nk);
     if pstr.Status=ok then
      val:=val + (val_nk/10.0**(pstr.pta-ptr));
      --write ("nk: "& str(val, 4));
     end if;
    end if;
    if sign=1 then  val:=-val; end if;
     --write("Vorzeichen:" & str(sign));
     --write ("ms: "& str(val, 4));
    if val >= wmin and val <= wmax then
     --write("Ausgabe");
     pstr.Status:=ok;
     w:=val;
     return;
    else
     pstr.Status:=err;
    end if;
   end if;
   pstr.pta:=ptr;
   pstr.err_pos:=ptr;
   assign(pstr.err_msg, "real range("&str(wmin)&" to "&str(wmax)&") expected");
  end if;
 end procedure;

-- Abraeumen einer Zeitangabe
  procedure get(pstr: inout tPString; w: out TIME; 
           wmin: TIME:=0 fs; wmax: TIME:= 1 sec) is
  constant ptr: POSITIVE:=pstr.pta;
  variable val: REAL;
  variable me: natural;
  variable t: time;
 begin
  if pstr.Status=ok and pstr.pta< pstr.pte then
   get(pstr, val);
   --write("TN1: "&str(pstr));

   if pstr.Status=ok then
    get_sym(pstr, me, " fs; ps; ns; us; ms; sec; min; hr;", ';');
    --write("TN2: "&str(pstr));
    if pstr.Status=ok then
     case me is
      when 0 => t:=val * fs;
      when 1 => t:=val * ps;
      when 2 => t:=val * ns;
      when 3 => t:=val * us;
      when 4 => t:=val * ms;
      when 5 => t:=val * sec;
      when 6 => t:=val * min;
      when others => t:=val * hr;
     end case;
     --write ("val, t: "& str(val)&" ," &str(t));
     if t>=wmin and t<=wmax then
      w:=t; return;
     else
      pstr.Status:=err;
     end if;
    end if;
   end if;
  pstr.pta:=ptr;
  pstr.err_pos:=ptr;
  assign(pstr.err_msg, "time range("&str(wmin)&" to "&str(wmax)&") expected");
  end if;
 end procedure;

---------------------------------------------
-- Bit
 procedure get(pstr: inout tPString; w: out BIT) is
  constant ptr: POSITIVE:=pstr.pta;
  variable pos: natural;
 begin
  if pstr.Status=ok then
   get_char(pstr, pos, "01");
   if pstr.Status=ok then
    w:=bit'val(pos);
    return;
   end if;
   pstr.pta:=ptr;
   pstr.err_pos:=ptr;
   assign(pstr.err_msg, "bit expected");
  end if;
 end procedure;

----------------------------------------------
-- BOOLEAN
 procedure get(pstr: inout tPString; w: out boolean) is
  constant ptr: POSITIVE:=pstr.pta;
  variable pos: natural;
 begin
  if pstr.Status=ok then
   get_sym(pstr, pos, "false true ");
   if pstr.Status=ok then
    w:=boolean'val(pos);
    return;
   end if;
   pstr.pta:=ptr;
   pstr.err_pos:=ptr;
   assign(pstr.err_msg, "boolean expected");
  end if;
 end procedure;

 procedure get(pstr: inout tPString; w: out std_ulogic) is
  constant ptr: POSITIVE:=pstr.pta;
  variable pos: natural;
 begin
  if pstr.Status=ok then
   get_char(pstr, pos, "UX01ZWLH-");
   if pstr.Status=ok then
    w:=std_ulogic'val(pos);
    return;
   end if;
   pstr.pta:=ptr;
   pstr.err_pos:=ptr;
   assign(pstr.err_msg, "std_(u)logic expected");
  end if;
 end procedure;

 procedure get(pstr: inout tPString; w: out bit_vector) is
  constant ptr: POSITIVE:=pstr.pta;
  variable pos: natural;
 begin
  if pstr.Status=ok then
   for idx in w'range loop
    get_char(pstr, pos, "01");
    --write("M1 pstr="&str(pstr)&" pos="&str(pos));
    if pstr.Status=ok and pstr.pta<=pstr.pte then
     w(idx):=bit'val(pos);
    else
     pstr.Status:=err;
     pstr.pta:=ptr;
     pstr.err_pos:=ptr;
     assign(pstr.err_msg, "bit_vector of length " & str(w'length) & " expected");
     return;
    end if;
   end loop;
  end if;
 end procedure;

 procedure get(pstr: inout tPString; w: out std_logic_vector) is
  constant ptr: POSITIVE:=pstr.pta;
  variable pos: natural;
 begin
  if pstr.Status=ok then
   for idx in w'range loop
    get_char(pstr, pos, "UX01ZWLH-");
    if pstr.Status=ok and pstr.pta<=pstr.pte then
     w(idx):=std_ulogic'val(pos);
    else
     pstr.Status:=err;
     pstr.pta:=ptr;
     pstr.err_pos:=ptr;
     assign(pstr.err_msg, "std_logic_vector of length " & str(w'length) & " expected");
     return;
    end if;
   end loop;
  end if;
 end procedure;

---------------------------------------------------------------
-- Dialogeingabe
-- Hilfsprozedur, die bei Eingabe eines Symbols aus dem ExitStr
-- Die Simulation beendet

 constant ExitStr: string:= "exit Exit stop Stop halt Halt ";
 procedure CheckExit(pstr: inout tPString) is
  variable pos: natural;
 begin
  skip(pstr, " ");
  get_sym(pstr, pos, ExitStr);
  if pstr.Status=ok then 
   StopSim("Simulation durch Eingabe von " & str(pos, ExitStr) & " beendet.");
  else
   pstr.Status:=ok;
  end if;
 end procedure;


 procedure read(file f: TEXT; pstr: out tPString) is
  variable vpstr: tPString;
 begin
  std.textio.read(f, vpstr.str, vpstr.pte);
  if vpstr.pte>vpstr.str'length+1 then
   vpstr.pte:=vpstr.str'length;
   vpstr.Status:=trunc;
  else
   --vpstr.pte:=vpstr.pte-1;
   vpstr.Status:=ok;
  end if;
  pstr:=vpstr;
 end procedure;


 procedure read(promt: string; w: out tPString) is
  variable pstr: tPString;
 begin
  write(promt);
  read(input, w);
 end procedure;

 procedure read(promt: string; w: out INTEGER; 
           wmin: INTEGER:=INTEGER'low+1; wmax: INTEGER:= INTEGER'high) is
  variable pstr: tPString;
 begin
  write(promt);
  loop
   read(input, pstr);
   CheckExit(pstr);
   get(pstr, w, wmin, wmax);
   exit when pstr.Status=ok;
   write(str(pstr.err_msg));
  end loop;
 end procedure; 


 procedure read(promt: string; w: out REAL; 
           wmin: REAL:=REAL'low/10.0; wmax: REAL:= REAL'high/10.0) is
   variable pstr: tPString;
  begin
  write(promt);
  loop
   read(input, pstr);
   CheckExit(pstr);
   get(pstr, w, wmin, wmax);
   exit when pstr.Status=ok;
   write(str(pstr.err_msg));
  end loop;
 end procedure; 


 procedure read(promt: string; w: out BIT) is
   variable pstr: tPString;
  begin
  write(promt);
  loop
   read(input, pstr);
   CheckExit(pstr);
   get(pstr, w);
   exit when pstr.Status=ok;
   write(str(pstr.err_msg));
  end loop;
 end procedure; 

 procedure read(promt: string; w: out boolean) is
   variable pstr: tPString;
  begin
  write(promt);
  loop
   read(input, pstr);
   CheckExit(pstr);
   get(pstr, w);
   exit when pstr.Status=ok;
   write(str(pstr.err_msg));
  end loop;
 end procedure; 

 procedure read(promt: string; w: out std_ulogic) is
   variable pstr: tPString;
  begin
  write(promt);
  loop
   read(input, pstr);
   CheckExit(pstr);
   get(pstr, w);
   exit when pstr.Status=ok;
   write(str(pstr.err_msg));
  end loop;
 end procedure; 


 procedure read(promt: string; w: out bit_vector) is
   variable pstr: tPString;
  begin
  write(promt);
  loop
   read(input, pstr);
   CheckExit(pstr);
   get(pstr, w);
   exit when pstr.Status=ok;
   write(str(pstr.err_msg));
  end loop;
 end procedure; 


 procedure read(promt: string; w: out std_logic_vector) is
   variable pstr: tPString;
  begin
  write(promt);
  loop
   read(input, pstr); 
   CheckExit(pstr);
   get(pstr, w);
   exit when pstr.Status=ok;
   write(str(pstr.err_msg));
  end loop;
 end procedure; 

-- Warte auf Eingabe ohne Auswertung des Wertes
 procedure read is
   variable pstr: tPString;
 begin
  write("Weiter mit Eingabetaste (Abbruch mit Exit)");
  read(input, pstr);
  CheckExit(pstr);
 end procedure;


end package body;

  

  
  


