----------------------------------------------------------------
-- G. Kemnitz: Technische Informatik 
-- Band 2: Entwurf digitaler Schaltungen
-- Bibliothek Tuc
----------------------------------------------------------------
-- Package Tuc.Numeric_Synth
--
-- entstanden aus Package Tuc.Numeric.Sim vom 15.02.2010 von G. Kemnitz
-- 
-- definiert die mit STD_LOGIC_VECTOR eng verwandten Datentypen
-- tSigned und tUnsigned zur Darstellung vorzeichenbehafteter und
-- vorzeichenfreier ganzer Zahlen und ueberlaedt fuer diese 
-- folgende Funktionen und Operatoren:
--
-- entfernt, da nicht synthetisierbar
-- is_x(z) 	=> BOOLEAN		Test auf Ungueltigkeit
-- str(z, fmt) 	=> STRING 		Erzeugung einer Textdarstellung
-- get(pstr, z)                         Einlesen des Wertes aus einem Parser-Objekt
-- read(promt, z)                       Einlesen des Wertes von der Standardeingabe 
-- beibehalten, aber ohne Abfrage auf X
-- int(z) 	=> INTEGER/NATURAL   	Umwandlung in ein Zahl
-- to_tSigned(i) => tSigned		Umwandlung einer Zahl in einen Bitvektor
-- to_tUnsigned(i) => tUnsigned		
-- abs a, -a				einstellige aritmetische Operatoren, nur tSigned
-- +, -, *, /, rem, mod			zweistellige arithmetische Operatoren
-- =, /=, >, <, >=, <= 			Vergleichsoperatoren
-- sll, srl, sra, rol, rot		Verschiebeoperatoren
-- not, and, nand, or, nor, xor, xnor	logische Operatoren
--
-- 31.03.2010, C. Giesemann
----------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_BIT.all;
library Tuc;
use Tuc.Ausgabe.all;
--use Tuc.Eingabe.all;

package Numeric_Synth is

  --============================================================================
  -- Numeric array type definitions
  --============================================================================

  type tSigned   is array (NATURAL range <> ) of STD_LOGIC;
  type tUnsigned is array (NATURAL range <> ) of STD_LOGIC;

  --function is_x(x: tSigned) return BOOLEAN;
  --function is_x(x: tUnsigned) return BOOLEAN;
  function str(x: tSigned; fmt:CHARACTER:='b') return STRING;
  function str(x: tUnsigned; fmt:CHARACTER:='b') return STRING;

  --procedure get(pstr: inout tPString; w: out tSigned);
  --procedure get(pstr: inout tPString; w: out tUnsigned);
  --procedure read(promt: string; w: out tSigned);
  --procedure read(promt: string; w: out tUnsigned);

  --============================================================================
  -- Conversion Functions
  --============================================================================

  function int(x: tSigned) return INTEGER;
  function int(x: tUnsigned) return INTEGER;
  function to_tUnsigned(x: INTEGER; n: NATURAL) return tUnsigned;
  function to_tSigned(x: INTEGER; n: NATURAL) return tSigned;

  -- Abbildung: ungueltig <=> INTEGER'LOW
  --
  -- tSigned, tUnsigned und STD_LOGIC_VECTOR sind eng verwandte Datentypen
  -- die Konvertierung erfolgt mit dem vorangestellten Bezeichner des Zieltyps
  --============================================================================
 -- einstellige Arithmetikoperationen, Indexbereich Ergebnis: x'LENGTH-1 downto 0
  function "abs"(x: tSigned) return tSigned;
  function "-"  (x: tSigned) return tSigned;

 -- Addition und Subtraktion, Indexbereich Ergebnis: max(a'LENGTH, b'LENGTH)-1 downto 0)
  function "+" (a, b: tSigned)   return tSigned;
  function "+" (a, b: tUnsigned) return tUnsigned;
  function "-" (a, b: tSigned)   return tSigned;
  function "-" (a, b: tUnsigned) return tUnsigned;

 -- Multiplikation, Indexbereich Ergebnis: a'LENGTH + b'LENGTH-1 downto 0
  function "*" (a, b: tSigned)   return tSigned;
  function "*" (a, b: tUnsigned) return tUnsigned;

 -- Division, Indexbereich Ergebnis: a'LENGTH-1 downto 0
 -- [bei einer Division durch Null wird das Ergebnis auf ungueltig gesetzt 
 -- und eine Warnung generiert]
  function "/" (a, b: tSigned)   return tSigned;
  function "/" (a, b: tUnsigned) return tUnsigned;

 -- Divisionsrest und Modulo, Indexbereich Ergebnis: b'LENGTH-1 downto 0
 -- [bei einer Division durch Null wird das Ergebnis auf ungueltig gesetzt 
 -- und eine Warnung generiert]
  function "rem" (a, b: tSigned)   return tSigned;
  function "rem" (a, b: tUnsigned) return tUnsigned;
  function "mod" (a, b: tSigned)   return tSigned;
  function "mod" (a, b: tUnsigned) return tUnsigned;

 -- Vergleichsoperatoren
  function "="  (a, b: tSigned) return BOOLEAN;
  function "/=" (a, b: tSigned) return BOOLEAN;
  function ">"  (a, b: tSigned) return BOOLEAN;
  function "<"  (a, b: tSigned) return BOOLEAN;
  function ">=" (a, b: tSigned) return BOOLEAN;
  function "<=" (a, b: tSigned) return BOOLEAN;

  function "="  (a, b: tUnsigned) return BOOLEAN;
  function "/=" (a, b: tUnsigned) return BOOLEAN;
  function ">"  (a, b: tUnsigned) return BOOLEAN;
  function "<"  (a, b: tUnsigned) return BOOLEAN;
  function ">=" (a, b: tUnsigned) return BOOLEAN;
  function "<=" (a, b: tUnsigned) return BOOLEAN;

 -- Rotations- und Schiebeoperationen, Indexbereich Ergebnis x'RANGE
 -- diese Operatoren sind nicht in VHDL 1076-1987 definiert und bei einer Nutzung
 -- unter  VHDL 1076-1987 auszukommentieren
  function "sll" (a: tSigned; n: NATURAL) return tSigned;
  function "srl" (a: tSigned; n: NATURAL) return tSigned;
  function "rol" (a: tSigned; n: NATURAL) return tSigned;
  function "ror" (a: tSigned; n: NATURAL) return tSigned;
  function "sla" (a: tSigned; n: NATURAL) return tSigned;
  function "sra" (a: tSigned; n: NATURAL) return tSigned;

  function "sll" (a: tUnsigned; n: NATURAL) return tUnsigned;
  function "srl" (a: tUnsigned; n: NATURAL) return tUnsigned;
  function "rol" (a: tUnsigned; n: NATURAL) return tUnsigned;
  function "ror" (a: tUnsigned; n: NATURAL) return tUnsigned;

-- logische Operatoren  
  function "not"  (a: tSigned)    return tSigned;
  function "and"  (a, b: tSigned) return tSigned;
  function "nand" (a, b: tSigned) return tSigned;
  function "or"   (a, b: tSigned) return tSigned;
  function "nor"  (a, b: tSigned) return tSigned;
  function "xor"  (a, b: tSigned) return tSigned;
  function "xnor" (a, b: tSigned) return tSigned;

  function "not"  (a: tUnsigned)    return tUnsigned;
  function "and"  (a, b: tUnsigned) return tUnsigned;
  function "nand" (a, b: tUnsigned) return tUnsigned;
  function "or"   (a, b: tUnsigned) return tUnsigned;
  function "nor"  (a, b: tUnsigned) return tUnsigned;
  function "xor"  (a, b: tUnsigned) return tUnsigned;
  function "xnor" (a, b: tUnsigned) return tUnsigned;


 end package;

package body Numeric_Synth is
--==========================================================
-- nicht exportierte Hilfsfunktionen
-- Maximum
--function max(a, b:INTEGER) return INTEGER is
-- begin
--  if a>=b then return a; 
--  else return b;
--  end if;
-- end function;

-- Konvertierung tSigned => IEEE.NUMERIC_BIT.SIGNED
 function to_signed(x: tSigned) return SIGNED is
 begin
  return SIGNED(TO_BITVECTOR(STD_LOGIC_VECTOR(x)));
 end function;

-- Konvertierung tUnsigned => IEEE.NUMERIC_BIT.UNSIGNED
 function to_unsigned(x: tUnsigned) return UNSIGNED is
 begin
  return UNSIGNED(TO_BITVECTOR(STD_LOGIC_VECTOR(x)));
 end function;

-- Konvertierung IEEE.NUMERIC_BIT.SIGNED => tSigned
 function to_tSigned(x: SIGNED) return tSigned is
 begin
  return tSigned(TO_STDLOGICVECTOR(BIT_VECTOR(x)));
 end function;

-- Konvertierung IEEE.NUMERIC_BIT.UNSIGNED => tSigned
 function to_tUnsigned(x: UNSIGNED) return tUnsigned is
 begin
  return tUnsigned(TO_STDLOGICVECTOR(BIT_VECTOR(x)));
 end function;

  --============================================================================

-- Umwandlung in eine Zeichenkettendarstellung
  function str(x: tSigned; fmt:CHARACTER:='b') return STRING is
  begin
   if fmt='t' then return "tSigned";
   elsif fmt='d' then return str(STD_LOGIC_VECTOR(x), 's');
   else return str(STD_LOGIC_VECTOR(x), fmt);
   end if;
  end function;

  function str(x: tUnsigned; fmt:CHARACTER:='b') return STRING is
  begin
   if fmt='t' then return "tUnsigned";
   elsif fmt='s' then return str(STD_LOGIC_VECTOR(x), 'd');
   else return str(STD_LOGIC_VECTOR(x), fmt);
   end if;
  end function;

-- Einlesen eines Wertes von einem Parser-Objekt
--  procedure get(pstr: inout tPString; w: out tSigned) is
--   variable slv: STD_LOGIC_VECTOR(w'RANGE);
--  begin
--   get(pstr, slv);
--   w:= tSigned(slv);
--  end procedure;

--  procedure get(pstr: inout tPString; w: out tUnsigned) is
--   variable slv: STD_LOGIC_VECTOR(w'RANGE);
--  begin
--   get(pstr, slv);
--   w:= tUnsigned(slv);
--  end procedure;

-- Einlesen eines Wertes von der Standardeingabe
--  procedure read(promt: string; w: out tSigned) is
--   variable slv: STD_LOGIC_VECTOR(w'RANGE);
--  begin
--   read(promt, slv);
--   w:= tSigned(slv);
--  end procedure;

--  procedure read(promt: string; w: out tUnsigned) is
--   variable slv: STD_LOGIC_VECTOR(w'RANGE);
--  begin
--   read(promt, slv);
--   w:= tUnsigned(slv);
--  end procedure;


  --============================================================================
  -- Conversion Functions
  --============================================================================

  function int(x: tSigned) return INTEGER is
  begin
   return TO_INTEGER(to_signed(x));
  end function;

  function int(x: tUnsigned) return INTEGER is
  begin
  -- if IS_X(x) or x'LENGTH > 31 then return INTEGER'LOW;
  -- else 
   return TO_INTEGER(to_unsigned(x));
  -- end if;
  end function;

  function to_tSigned(x: INTEGER; n: NATURAL) return tSigned is
  -- constant cXX: tSigned(n-1 downto 0):= (others=>'X');
  begin
  -- if x=INTEGER'LOW then return cXX;
  -- else 
   return to_tSigned(TO_SIGNED(x, n));
  -- end if;
  end function;
 
  function to_tUnsigned(x: INTEGER; n: NATURAL) return tUnsigned is
  -- constant cXX: tUnsigned(n-1 downto 0):= (others=>'X');
  begin
  -- if x=INTEGER'LOW then return cXX;
  -- else 
   return to_tUnsigned(TO_UNSIGNED(x, n));
  -- end if;
  end function;

  --============================================================================
  -- einstellige Arithmetikoperationen
  -- Indexbereich Ergebnis: x'LENGTH-1 downto 0
  --============================================================================

  function "abs" (x: tSigned) return tSigned is
  -- constant cXX: tSigned(x'RANGE) := (others=>'X');
  begin
  -- if IS_X(x) then return cXX;
  -- else            
   return to_tSigned(abs(to_signed(x)));
  -- end if;
  end function;

  function "-" (x: tSigned) return tSigned is
  -- constant cXX: tSigned(x'RANGE) := (others=>'X');
  begin
  -- if IS_X(x) then return cXX;
  -- else            
    return to_tSigned(-(to_signed(x)));
  -- end if;
  end function;

  --============================================================================
  -- Addition und Subtraktion
  -- Indexbereich Ergebnis: max(a'LENGTH, b'LENGTH)-1 downto 0)
  --============================================================================

  function "+" (a, b: tSigned) return tSigned is
  -- constant cXX: tSigned(max(a'LENGTH, b'LENGTH)-1 downto 0) := (others=>'X');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- else 
   return to_tSigned(to_signed(a)+to_signed(b));
  -- end if;
  end function;

  function "+" (a, b: tUnsigned) return tUnsigned is
  -- constant cXX: tUnsigned(max(a'LENGTH, b'LENGTH)-1 downto 0) := (others=>'X');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- else 
   return to_tUnsigned(to_unsigned(a)+to_unsigned(b));
  -- end if;
  end function;

  function "-" (a, b: tSigned) return tSigned is
  -- constant cXX: tSigned(max(a'LENGTH, b'LENGTH)-1 downto 0) := (others=>'X');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- else 
   return to_tSigned(to_signed(a)-to_signed(b));
  -- end if;
  end function;

  function "-" (a, b: tUnsigned) return tUnsigned is
  -- constant cXX: tUnsigned(max(a'LENGTH, b'LENGTH)-1 downto 0) := (others=>'X');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- else 
   return to_tUnsigned(to_unsigned(a)-to_unsigned(b));
  -- end if;
  end function;

  --============================================================================
  -- Multiplikation
  -- Indexbereich Ergebnis: a'LENGTH + b'LENGTH-1 downto 0
  --============================================================================
  
  function "*" (a, b: tSigned) return tSigned is
  -- constant cXX: tSigned(a'LENGTH+b'LENGTH-1 downto 0) := (others=>'X');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- else 
   return to_tSigned(to_signed(a)*to_signed(b));
  -- end if;
  end function;
  
  function "*" (a, b: tUnsigned) return tUnsigned is
  -- constant cXX: tUnsigned(a'LENGTH+b'LENGTH-2 downto 0) := (others=>'X');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- else 
   return to_tUnsigned(to_unsigned(a)*to_unsigned(b));
  -- end if;
  end function;

  --============================================================================
  -- Division
  -- Indexbereich Ergebnis: a'LENGTH-1 downto 0; [bei einer Division durch Null
  -- wird das Ergebnis auf ungueltig gesetzt und eine Warnung generiert]
  --============================================================================
  function "/" (a, b: tSigned) return tSigned is
  -- constant cXX: tSigned(a'LENGTH-1 downto 0) := (others=>'X');
   constant cb0: tSigned(b'LENGTH-1 downto 0) := (others=>'0');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- elsif b/=cb0 then return to_tSigned(to_signed(a)/to_signed(b));
  -- else report "Division durch null" severity WARNING; return cXX;
   return to_tSigned(to_signed(a)/to_signed(b));
  end function;

  function "/" (a, b: tUnsigned) return tUnsigned is
  -- constant cXX: tUnsigned(a'LENGTH-1 downto 0) := (others=>'X');
   constant cb0: tUnsigned(b'LENGTH-1 downto 0) := (others=>'0');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- elsif b/=cb0 then return to_tUnsigned(to_unsigned(a)/to_unsigned(b));
  -- else report "Division durch null" severity WARNING; return cXX;
   return to_tUnsigned(to_unsigned(a)/to_unsigned(b));
  end function;

  --============================================================================
  -- Divisionsrest und Modulo; Indexbereich Ergebnis: b'LENGTH-1 downto 0
  -- [bei einer Division durch Null wird das Ergebnis auf ungueltig gesetzt 
  -- und eine Warnung generiert]
  --============================================================================

  function "rem" (a, b: tSigned) return tSigned is
  -- constant cXX: tSigned(b'LENGTH-1 downto 0) := (others=>'X');
   constant cb0: tSigned(b'LENGTH-1 downto 0) := (others=>'0');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- elsif b/=cb0 then return to_tSigned(to_signed(a) rem to_signed(b));
  -- else report "Division durch null" severity WARNING; return cXX;
   return to_tSigned(to_signed(a) rem to_signed(b));
  end function;

  function "rem" (a, b: tUnsigned) return tUnsigned is
   constant cXX: tUnsigned(b'LENGTH-1 downto 0) := (others=>'X');
   constant cb0: tUnsigned(b'LENGTH-1 downto 0) := (others=>'0');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- elsif b/=cb0 then return to_tUnsigned(to_unsigned(a) rem to_unsigned(b));
  -- else report "Division durch null" severity WARNING; return cXX;
   return to_tUnsigned(to_unsigned(a) rem to_unsigned(b));
  end function;

  function "mod" (a, b: tSigned) return tSigned is
   constant cXX: tSigned(b'LENGTH-1 downto 0) := (others=>'X');
   constant cb0: tSigned(b'LENGTH-1 downto 0) := (others=>'0');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- elsif b/=cb0 then return to_tSigned(to_signed(a) mod to_signed(b));
  -- else report "Division durch null" severity WARNING; return cXX;
   return to_tSigned(to_signed(a) mod to_signed(b));
  end function;

  function "mod" (a, b: tUnsigned) return tUnsigned is
   constant cXX: tUnsigned(b'LENGTH-1 downto 0) := (others=>'X');
   constant cb0: tUnsigned(b'LENGTH-1 downto 0) := (others=>'0');
  begin
  -- if is_x(a) or is_x(b) then return cXX;
  -- elsif b/=cb0 then return to_tUnsigned(to_unsigned(a) mod to_unsigned(b));
  -- else report "Division durch null" severity WARNING; return cXX;
   return to_tUnsigned(to_unsigned(a) mod to_unsigned(b));
  end function;

 
-- ============================================================
-- Vergleichsoperatoren
-- [fuer ungueltige Operanden immer False]
-- ============================================================

  function "="  (a, b: tSigned) return BOOLEAN is
  begin
  -- return (not is_x(a)) and (not is_x(b)) and STD_LOGIC_VECTOR(a) = STD_LOGIC_VECTOR(b);
   return STD_LOGIC_VECTOR(a) = STD_LOGIC_VECTOR(b);
  end function;

  function "/="  (a, b: tSigned) return BOOLEAN is
  begin
  -- return (not is_x(a)) and (not is_x(b)) and STD_LOGIC_VECTOR(a) /= STD_LOGIC_VECTOR(b);
   return STD_LOGIC_VECTOR(a) /= STD_LOGIC_VECTOR(b);
  end function;

  function ">"  (a, b: tSigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_signed(a)>to_signed(b);
  -- end if;
  end function;

  function "<"  (a, b: tSigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_signed(a)<to_signed(b);
  -- end if;
  end function;

  function ">="  (a, b: tSigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_signed(a)>=to_signed(b);
  -- end if;
  end function;

  function "<="  (a, b: tSigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_signed(a)<=to_signed(b);
  -- end if;
  end function;

   ---------------------------------------------------------

  function "="  (a, b: tUnsigned) return BOOLEAN is
  begin
  -- return (not is_x(a)) and (not is_x(b)) and STD_LOGIC_VECTOR(a) = STD_LOGIC_VECTOR(b);
   return STD_LOGIC_VECTOR(a) = STD_LOGIC_VECTOR(b);
  end function;

  function "/="  (a, b: tUnsigned) return BOOLEAN is
  begin
  -- return (not is_x(a)) and (not is_x(b)) and STD_LOGIC_VECTOR(a) /= STD_LOGIC_VECTOR(b);
   return STD_LOGIC_VECTOR(a) /= STD_LOGIC_VECTOR(b);
  end function;

  function ">"  (a, b: tUnsigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_unsigned(a)>to_unsigned(b);
  -- end if;
  end function;

  function "<"  (a, b: tUnsigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_unsigned(a)<to_unsigned(b);
  -- end if;
  end function;

  function ">="  (a, b: tUnsigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_unsigned(a)>=to_unsigned(b);
  -- end if;
  end function;

  function "<="  (a, b: tUnsigned) return BOOLEAN is
  begin
  -- if is_x(a) or is_X(b) then return FALSE;
  -- else 
   return to_unsigned(a)<=to_unsigned(b);
  -- end if;
  end function;

 -- ============================================================
 -- Rotations- und Schiebeoperationen, Indexbereich Ergebnis x'RANGE
 -- diese Operatoren sind nicht in VHDL 1076-1987 definiert und bei einer Nutzung
 -- unter  VHDL 1076-1987 auszukommentieren
 -- ============================================================

  function "sll" (a: tSigned; n: NATURAL) return tSigned is
  begin
   return to_tSigned(to_signed(a) sll n);
  end function;

  function "srl" (a: tSigned; n: NATURAL) return tSigned is
  begin
   return to_tSigned(to_signed(a) srl n);
  end function;

  function "rol" (a: tSigned; n: NATURAL) return tSigned is
  begin
   return to_tSigned(to_signed(a) rol n);
  end function;

  function "ror" (a: tSigned; n: NATURAL) return tSigned is
  begin
   return to_tSigned(to_signed(a) ror n);
  end function;

  function "sla" (a: tSigned; n: NATURAL) return tSigned is
  begin
   return to_tSigned(to_signed(a) sla n);
  end function;

  function "sra" (a: tSigned; n: NATURAL) return tSigned is
  begin
   return to_tSigned(to_signed(a) sra n);
  end function;
 --------------------------------------------------------------------

  function "sll" (a: tUnsigned; n: NATURAL) return tUnsigned is
  begin
   return to_tUnsigned(to_unsigned(a) sll n);
  end function;

  function "srl" (a: tUnsigned; n: NATURAL) return tUnsigned is
  begin
   return to_tUnsigned(to_unsigned(a) srl n);
  end function;

  function "rol" (a: tUnsigned; n: NATURAL) return tUnsigned is
  begin
   return to_tUnsigned(to_unsigned(a) rol n);
  end function;

  function "ror" (a: tUnsigned; n: NATURAL) return tUnsigned is
  begin
   return to_tUnsigned(to_unsigned(a) ror n);
  end function;

-- ============================================================
-- logische Operatoren  

  function "not"  (a: tSigned) return tSigned is
  begin
   return tSigned(not STD_LOGIC_VECTOR(a));
  end function;

  function "and"  (a, b: tSigned) return tSigned is
  begin
   return tSigned(STD_LOGIC_VECTOR(a) and STD_LOGIC_VECTOR(b));
  end function;

  function "nand" (a, b: tSigned) return tSigned is
  begin
   return tSigned(STD_LOGIC_VECTOR(a) nand STD_LOGIC_VECTOR(b));
  end function;

  function "or"   (a, b: tSigned) return tSigned is
  begin
   return tSigned(STD_LOGIC_VECTOR(a) or STD_LOGIC_VECTOR(b));
  end function;

  function "nor"  (a, b: tSigned) return tSigned is
  begin
   return tSigned(STD_LOGIC_VECTOR(a) nor STD_LOGIC_VECTOR(b));
  end function;

  function "xor"  (a, b: tSigned) return tSigned is
  begin
   return tSigned(STD_LOGIC_VECTOR(a) xor STD_LOGIC_VECTOR(b));
  end function;

  function "xnor" (a, b: tSigned) return tSigned is
  begin
   return tSigned(STD_LOGIC_VECTOR(a) xnor STD_LOGIC_VECTOR(b));
  end function;

  function "not"  (a: tUnsigned) return tUnsigned is
  begin
   return tUnsigned(not STD_LOGIC_VECTOR(a));
  end function;

  function "and"  (a, b: tUnsigned) return tUnsigned is
  begin
   return tUnsigned(STD_LOGIC_VECTOR(a) and STD_LOGIC_VECTOR(b));
  end function;

  function "nand" (a, b: tUnsigned) return tUnsigned is
  begin
   return tUnsigned(STD_LOGIC_VECTOR(a) nand STD_LOGIC_VECTOR(b));
  end function;

  function "or"   (a, b: tUnsigned) return tUnsigned is
  begin
   return tUnsigned(STD_LOGIC_VECTOR(a) or STD_LOGIC_VECTOR(b));
  end function;

  function "nor"  (a, b: tUnsigned) return tUnsigned is
  begin
   return tUnsigned(STD_LOGIC_VECTOR(a) nor STD_LOGIC_VECTOR(b));
  end function;

  function "xor"  (a, b: tUnsigned) return tUnsigned is
  begin
   return tUnsigned(STD_LOGIC_VECTOR(a) xor STD_LOGIC_VECTOR(b));
  end function;

  function "xnor" (a, b: tUnsigned) return tUnsigned is
  begin
   return tUnsigned(STD_LOGIC_VECTOR(a) xnor STD_LOGIC_VECTOR(b));
  end function;


 end package body;
